'''
Created on Jul 2, 2014

@author: roj-idl71
'''

import numpy

from .jroIO_base import LOCALTIME, JRODataReader, JRODataWriter
from schainpy.model.proc.jroproc_base import ProcessingUnit, Operation, MPDecorator
from schainpy.model.data.jroheaderIO import PROCFLAG, BasicHeader, SystemHeader, RadarControllerHeader, ProcessingHeader
from schainpy.model.data.jrodata import Voltage


class VoltageReader(JRODataReader, ProcessingUnit):
    """
    Esta clase permite leer datos de voltage desde archivos en formato rawdata (.r). La lectura
    de los datos siempre se realiza por bloques. Los datos leidos (array de 3 dimensiones:
    perfiles*alturas*canales) son almacenados en la variable "buffer".

                             perfiles * alturas * canales

    Esta clase contiene instancias (objetos) de las clases BasicHeader, SystemHeader,
    RadarControllerHeader y Voltage. Los tres primeros se usan para almacenar informacion de la
    cabecera de datos (metadata), y el cuarto (Voltage) para obtener y almacenar un perfil de
    datos desde el "buffer" cada vez que se ejecute el metodo "getData".

    Example:

        dpath = "/home/myuser/data"

        startTime = datetime.datetime(2010,1,20,0,0,0,0,0,0)

        endTime = datetime.datetime(2010,1,21,23,59,59,0,0,0)

        readerObj = VoltageReader()

        readerObj.setup(dpath, startTime, endTime)

        while(True):

            #to get one profile
            profile =  readerObj.getData()

            #print the profile
            print profile

            #If you want to see all datablock
            print readerObj.datablock

            if readerObj.flagNoMoreFiles:
                break

    """

    def __init__(self):
        """
        Inicializador de la clase VoltageReader para la lectura de datos de voltage.

        Input:
            dataOut    :    Objeto de la clase Voltage. Este objeto sera utilizado para
                              almacenar un perfil de datos cada vez que se haga un requerimiento
                              (getData). El perfil sera obtenido a partir del buffer de datos,
                              si el buffer esta vacio se hara un nuevo proceso de lectura de un
                              bloque de datos.
                              Si este parametro no es pasado se creara uno internamente.

        Variables afectadas:
            self.dataOut

        Return:
            None
        """

        ProcessingUnit.__init__(self)
        
        self.ext = ".r"
        self.optchar = "D"
        self.basicHeaderObj = BasicHeader(LOCALTIME)
        self.systemHeaderObj = SystemHeader()
        self.radarControllerHeaderObj = RadarControllerHeader()
        self.processingHeaderObj = ProcessingHeader()
        self.lastUTTime = 0
        self.profileIndex = 2**32 - 1        
        self.dataOut = Voltage()
        self.selBlocksize = None
        self.selBlocktime = None

    def createObjByDefault(self):

        dataObj = Voltage()

        return dataObj

    def __hasNotDataInBuffer(self):

        if self.profileIndex >= self.processingHeaderObj.profilesPerBlock * self.nTxs:
            return 1

        return 0

    def getBlockDimension(self):
        """
            Obtiene la cantidad de puntos a leer por cada bloque de datos

            Affected:
                self.blocksize

            Return:
                None
        """
        pts2read = self.processingHeaderObj.profilesPerBlock * \
            self.processingHeaderObj.nHeights * self.systemHeaderObj.nChannels
        self.blocksize = pts2read

    def readBlock(self):
        """
            readBlock lee el bloque de datos desde la posicion actual del puntero del archivo
            (self.fp) y actualiza todos los parametros relacionados al bloque de datos
            (metadata + data). La data leida es almacenada en el buffer y el contador del buffer
            es seteado a 0

            Inputs:
                None

            Return:
                None

            Affected:
                self.profileIndex
                self.datablock
                self.flagIsNewFile
                self.flagIsNewBlock
                self.nTotalBlocks

            Exceptions: 
                Si un bloque leido no es un bloque valido
        """

        # if self.server is not None:
        #     self.zBlock = self.receiver.recv()
        #     self.zHeader = self.zBlock[:24]
        #     self.zDataBlock = self.zBlock[24:]
        #     junk = numpy.fromstring(self.zDataBlock, numpy.dtype([('real','<i4'),('imag','<i4')]))
        #     self.processingHeaderObj.profilesPerBlock = 240
        #     self.processingHeaderObj.nHeights = 248
        #     self.systemHeaderObj.nChannels
        # else:
        current_pointer_location = self.fp.tell()
        junk = numpy.fromfile(self.fp, self.dtype, self.blocksize)

        try:
            junk = junk.reshape((self.processingHeaderObj.profilesPerBlock,
                                 self.processingHeaderObj.nHeights, self.systemHeaderObj.nChannels))
        except:
            # print "The read block (%3d) has not enough data" %self.nReadBlocks

            if self.waitDataBlock(pointer_location=current_pointer_location):
                junk = numpy.fromfile(self.fp, self.dtype, self.blocksize)
                junk = junk.reshape((self.processingHeaderObj.profilesPerBlock,
                                     self.processingHeaderObj.nHeights, self.systemHeaderObj.nChannels))
        #             return 0

        # Dimensions : nChannels, nProfiles, nSamples

        junk = numpy.transpose(junk, (2, 0, 1))
        self.datablock = junk['real'] + junk['imag'] * 1j

        self.profileIndex = 0

        self.flagIsNewFile = 0
        self.flagIsNewBlock = 1

        self.nTotalBlocks += 1
        self.nReadBlocks += 1

        return 1

    def getFirstHeader(self):

        self.getBasicHeader()

        self.dataOut.processingHeaderObj = self.processingHeaderObj.copy()

        self.dataOut.systemHeaderObj = self.systemHeaderObj.copy()

        self.dataOut.radarControllerHeaderObj = self.radarControllerHeaderObj.copy()

        if self.nTxs > 1:
            self.dataOut.radarControllerHeaderObj.ippSeconds = self.radarControllerHeaderObj.ippSeconds / self.nTxs
        # Time interval and code are propierties of dataOut. Its value depends of radarControllerHeaderObj.

        #         self.dataOut.timeInterval = self.radarControllerHeaderObj.ippSeconds * self.processingHeaderObj.nCohInt
        #
        #         if self.radarControllerHeaderObj.code is not None:
        #
        #             self.dataOut.nCode = self.radarControllerHeaderObj.nCode
        #
        #             self.dataOut.nBaud = self.radarControllerHeaderObj.nBaud
        #
        #             self.dataOut.code = self.radarControllerHeaderObj.code

        self.dataOut.dtype = self.dtype

        self.dataOut.nProfiles = self.processingHeaderObj.profilesPerBlock

        self.dataOut.heightList = numpy.arange(
            self.processingHeaderObj.nHeights) * self.processingHeaderObj.deltaHeight + self.processingHeaderObj.firstHeight

        self.dataOut.channelList = list(range(self.systemHeaderObj.nChannels))

        self.dataOut.nCohInt = self.processingHeaderObj.nCohInt

        # asumo q la data no esta decodificada
        self.dataOut.flagDecodeData = self.processingHeaderObj.flag_decode

        # asumo q la data no esta sin flip
        self.dataOut.flagDeflipData = self.processingHeaderObj.flag_deflip

        self.dataOut.flagShiftFFT = self.processingHeaderObj.shif_fft

    def reshapeData(self):

        if self.nTxs < 0:
            return

        if self.nTxs == 1:
            return

        if self.nTxs < 1 and self.processingHeaderObj.profilesPerBlock % (1. / self.nTxs) != 0:
            raise ValueError("1./nTxs (=%f), should be a multiple of nProfiles (=%d)" % (
                1. / self.nTxs, self.processingHeaderObj.profilesPerBlock))

        if self.nTxs > 1 and self.processingHeaderObj.nHeights % self.nTxs != 0:
            raise ValueError("nTxs (=%d), should be a multiple of nHeights (=%d)" % (
                self.nTxs, self.processingHeaderObj.nHeights))

        self.datablock = self.datablock.reshape(
            (self.systemHeaderObj.nChannels, self.processingHeaderObj.profilesPerBlock * self.nTxs, int(self.processingHeaderObj.nHeights / self.nTxs)))

        self.dataOut.nProfiles = self.processingHeaderObj.profilesPerBlock * self.nTxs
        self.dataOut.heightList = numpy.arange(self.processingHeaderObj.nHeights / self.nTxs) * \
            self.processingHeaderObj.deltaHeight + self.processingHeaderObj.firstHeight
        self.dataOut.radarControllerHeaderObj.ippSeconds = self.radarControllerHeaderObj.ippSeconds / self.nTxs

        return

    def readFirstHeaderFromServer(self):

        self.getFirstHeader()

        self.firstHeaderSize = self.basicHeaderObj.size

        datatype = int(numpy.log2((self.processingHeaderObj.processFlags &
                                   PROCFLAG.DATATYPE_MASK)) - numpy.log2(PROCFLAG.DATATYPE_CHAR))
        if datatype == 0:
            datatype_str = numpy.dtype([('real', '<i1'), ('imag', '<i1')])
        elif datatype == 1:
            datatype_str = numpy.dtype([('real', '<i2'), ('imag', '<i2')])
        elif datatype == 2:
            datatype_str = numpy.dtype([('real', '<i4'), ('imag', '<i4')])
        elif datatype == 3:
            datatype_str = numpy.dtype([('real', '<i8'), ('imag', '<i8')])
        elif datatype == 4:
            datatype_str = numpy.dtype([('real', '<f4'), ('imag', '<f4')])
        elif datatype == 5:
            datatype_str = numpy.dtype([('real', '<f8'), ('imag', '<f8')])
        else:
            raise ValueError('Data type was not defined')

        self.dtype = datatype_str
        #self.ippSeconds = 2 * 1000 * self.radarControllerHeaderObj.ipp / self.c
        self.fileSizeByHeader = self.processingHeaderObj.dataBlocksPerFile * self.processingHeaderObj.blockSize + \
            self.firstHeaderSize + self.basicHeaderSize * \
            (self.processingHeaderObj.dataBlocksPerFile - 1)
        #        self.dataOut.channelList = numpy.arange(self.systemHeaderObj.numChannels)
        #        self.dataOut.channelIndexList = numpy.arange(self.systemHeaderObj.numChannels)
        self.getBlockDimension()

    def getFromServer(self):
        self.flagDiscontinuousBlock = 0
        self.profileIndex = 0
        self.flagIsNewBlock = 1
        self.dataOut.flagNoData = False
        self.nTotalBlocks += 1
        self.nReadBlocks += 1
        self.blockPointer = 0

        block = self.receiver.recv()

        self.basicHeaderObj.read(block[self.blockPointer:])
        self.blockPointer += self.basicHeaderObj.length
        self.systemHeaderObj.read(block[self.blockPointer:])
        self.blockPointer += self.systemHeaderObj.length
        self.radarControllerHeaderObj.read(block[self.blockPointer:])
        self.blockPointer += self.radarControllerHeaderObj.length
        self.processingHeaderObj.read(block[self.blockPointer:])
        self.blockPointer += self.processingHeaderObj.length
        self.readFirstHeaderFromServer()

        timestamp = self.basicHeaderObj.get_datatime()
        print('[Reading] - Block {} - {}'.format(self.nTotalBlocks, timestamp))
        current_pointer_location = self.blockPointer
        junk = numpy.fromstring(
            block[self.blockPointer:], self.dtype, self.blocksize)

        try:
            junk = junk.reshape((self.processingHeaderObj.profilesPerBlock,
                                 self.processingHeaderObj.nHeights, self.systemHeaderObj.nChannels))
        except:
            # print "The read block (%3d) has not enough data" %self.nReadBlocks
            if self.waitDataBlock(pointer_location=current_pointer_location):
                junk = numpy.fromstring(
                    block[self.blockPointer:], self.dtype, self.blocksize)
                junk = junk.reshape((self.processingHeaderObj.profilesPerBlock,
                                     self.processingHeaderObj.nHeights, self.systemHeaderObj.nChannels))
        #               return 0

        # Dimensions : nChannels, nProfiles, nSamples

        junk = numpy.transpose(junk, (2, 0, 1))
        self.datablock = junk['real'] + junk['imag'] * 1j
        self.profileIndex = 0
        if self.selBlocksize == None:
            self.selBlocksize = self.dataOut.nProfiles
        if self.selBlocktime != None:
            if self.dataOut.nCohInt is not None:
                nCohInt = self.dataOut.nCohInt
            else:
                nCohInt = 1
            self.selBlocksize = int(self.dataOut.nProfiles * round(self.selBlocktime / (
                nCohInt * self.dataOut.ippSeconds * self.dataOut.nProfiles)))
        self.dataOut.data = self.datablock[:,
                                           self.profileIndex:self.profileIndex + self.selBlocksize, :]
        datasize = self.dataOut.data.shape[1]
        if datasize < self.selBlocksize:
            buffer = numpy.zeros(
                (self.dataOut.data.shape[0], self.selBlocksize, self.dataOut.data.shape[2]), dtype='complex')
            buffer[:, :datasize, :] = self.dataOut.data
            self.dataOut.data = buffer
            self.profileIndex = blockIndex

        self.dataOut.flagDataAsBlock = True
        self.flagIsNewBlock = 1
        self.dataOut.realtime = self.online

        return self.dataOut.data

    def getData(self):
        """
            getData obtiene una unidad de datos del buffer de lectura, un perfil,  y la copia al objeto self.dataOut
            del tipo "Voltage" con todos los parametros asociados a este (metadata). cuando no hay datos
            en el buffer de lectura es necesario hacer una nueva lectura de los bloques de datos usando
            "readNextBlock"

            Ademas incrementa el contador del buffer "self.profileIndex" en 1.

            Return:

                Si el flag self.getByBlock ha sido seteado el bloque completo es copiado a self.dataOut y el self.profileIndex
                es igual al total de perfiles leidos desde el archivo.

                Si self.getByBlock == False:

                    self.dataOut.data = buffer[:, thisProfile, :]

                    shape = [nChannels, nHeis]

                Si self.getByBlock == True:

                    self.dataOut.data = buffer[:, :, :]

                    shape = [nChannels, nProfiles, nHeis]

            Variables afectadas:
                self.dataOut
                self.profileIndex

            Affected:
                self.dataOut
                self.profileIndex
                self.flagDiscontinuousBlock
                self.flagIsNewBlock
        """
        if self.flagNoMoreFiles:
            self.dataOut.flagNoData = True
            return 0
        self.flagDiscontinuousBlock = 0
        self.flagIsNewBlock = 0
        if self.__hasNotDataInBuffer():
            if not(self.readNextBlock()):
                return 0

            self.getFirstHeader()

            self.reshapeData()
        if self.datablock is None:
            self.dataOut.flagNoData = True
            return 0

        if not self.getByBlock:

            """
                Return profile by profile

                If nTxs > 1 then one profile is divided by nTxs and number of total
                blocks is increased by nTxs (nProfiles *= nTxs)
            """
            self.dataOut.flagDataAsBlock = False
            self.dataOut.data = self.datablock[:, self.profileIndex, :]
            self.dataOut.profileIndex = self.profileIndex

            self.profileIndex += 1

        else:
            """
                Return a block
            """
            if self.selBlocksize == None:
                self.selBlocksize = self.dataOut.nProfiles
            if self.selBlocktime != None:
                if self.dataOut.nCohInt is not None:
                    nCohInt = self.dataOut.nCohInt
                else:
                    nCohInt = 1
                self.selBlocksize = int(self.dataOut.nProfiles * round(self.selBlocktime / (
                    nCohInt * self.dataOut.ippSeconds * self.dataOut.nProfiles)))

            self.dataOut.data = self.datablock[:,
                                               self.profileIndex:self.profileIndex + self.selBlocksize, :]
            self.profileIndex += self.selBlocksize
            datasize = self.dataOut.data.shape[1]

            if datasize < self.selBlocksize:
                buffer = numpy.zeros(
                    (self.dataOut.data.shape[0], self.selBlocksize, self.dataOut.data.shape[2]), dtype='complex')
                buffer[:, :datasize, :] = self.dataOut.data

                while datasize < self.selBlocksize:  # Not enough profiles to fill the block
                    if not(self.readNextBlock()):
                        return 0
                    self.getFirstHeader()
                    self.reshapeData()
                    if self.datablock is None:
                        self.dataOut.flagNoData = True
                        return 0
                    # stack data
                    blockIndex = self.selBlocksize - datasize
                    datablock1 = self.datablock[:, :blockIndex, :]

                    buffer[:, datasize:datasize +
                           datablock1.shape[1], :] = datablock1
                    datasize += datablock1.shape[1]

                self.dataOut.data = buffer
                self.profileIndex = blockIndex

            self.dataOut.flagDataAsBlock = True
            self.dataOut.nProfiles = self.dataOut.data.shape[1]

        self.dataOut.flagNoData = False

        self.getBasicHeader()

        self.dataOut.realtime = self.online

        return self.dataOut.data


@MPDecorator
class VoltageWriter(JRODataWriter, Operation):
    """
    Esta clase permite escribir datos de voltajes a archivos procesados (.r). La escritura
    de los datos siempre se realiza por bloques.
    """

    ext = ".r"

    optchar = "D"

    shapeBuffer = None

    def __init__(self):#, **kwargs):
        """
        Inicializador de la clase VoltageWriter para la escritura de datos de espectros.

        Affected:
            self.dataOut

        Return: None
        """
        Operation.__init__(self)#, **kwargs)

        self.nTotalBlocks = 0

        self.profileIndex = 0

        self.isConfig = False

        self.fp = None

        self.flagIsNewFile = 1

        self.blockIndex = 0

        self.flagIsNewBlock = 0

        self.setFile = None

        self.dtype = None

        self.path = None

        self.filename = None

        self.basicHeaderObj = BasicHeader(LOCALTIME)

        self.systemHeaderObj = SystemHeader()

        self.radarControllerHeaderObj = RadarControllerHeader()

        self.processingHeaderObj = ProcessingHeader()

    def hasAllDataInBuffer(self):
        if self.profileIndex >= self.processingHeaderObj.profilesPerBlock:
            return 1
        return 0

    def setBlockDimension(self):
        """
        Obtiene las formas dimensionales del los subbloques de datos que componen un bloque

        Affected:
            self.shape_spc_Buffer
            self.shape_cspc_Buffer
            self.shape_dc_Buffer

        Return: None
        """
        self.shapeBuffer = (self.processingHeaderObj.profilesPerBlock,
                            self.processingHeaderObj.nHeights,
                            self.systemHeaderObj.nChannels)

        self.datablock = numpy.zeros((self.systemHeaderObj.nChannels,
                                      self.processingHeaderObj.profilesPerBlock,
                                      self.processingHeaderObj.nHeights),
                                     dtype=numpy.dtype('complex64'))

    def writeBlock(self):
        """
        Escribe el buffer en el file designado

        Affected:
            self.profileIndex
            self.flagIsNewFile
            self.flagIsNewBlock
            self.nTotalBlocks
            self.blockIndex

        Return: None
        """
        data = numpy.zeros(self.shapeBuffer, self.dtype)

        junk = numpy.transpose(self.datablock, (1, 2, 0))

        data['real'] = junk.real
        data['imag'] = junk.imag

        data = data.reshape((-1))

        data.tofile(self.fp)

        self.datablock.fill(0)

        self.profileIndex = 0
        self.flagIsNewFile = 0
        self.flagIsNewBlock = 1

        self.blockIndex += 1
        self.nTotalBlocks += 1

#         print "[Writing] Block = %04d" %self.blockIndex

    def putData(self):
        """
        Setea un bloque de datos y luego los escribe en un file

        Affected:
            self.flagIsNewBlock
            self.profileIndex

        Return:
            0    :    Si no hay data o no hay mas files que puedan escribirse
            1    :    Si se escribio la data de un bloque en un file
        """
        if self.dataOut.flagNoData:
            return 0

        self.flagIsNewBlock = 0

        if self.dataOut.flagDiscontinuousBlock:
            self.datablock.fill(0)
            self.profileIndex = 0
            self.setNextFile()

        if self.profileIndex == 0:
            self.setBasicHeader()

        self.datablock[:, self.profileIndex, :] = self.dataOut.data

        self.profileIndex += 1

        if self.hasAllDataInBuffer():
            # if self.flagIsNewFile:
            self.writeNextBlock()
#            self.setFirstHeader()

        return 1

    def __getBlockSize(self):
        '''
        Este metodos determina el cantidad de bytes para un bloque de datos de tipo Voltage
        '''

        dtype_width = self.getDtypeWidth()

        blocksize = int(self.dataOut.nHeights * self.dataOut.nChannels *
                        self.profilesPerBlock * dtype_width * 2)

        return blocksize

    def setFirstHeader(self):
        """
        Obtiene una copia del First Header

        Affected:
            self.systemHeaderObj
            self.radarControllerHeaderObj
            self.dtype

        Return:
            None
        """

        self.systemHeaderObj = self.dataOut.systemHeaderObj.copy()
        self.systemHeaderObj.nChannels = self.dataOut.nChannels
        self.radarControllerHeaderObj = self.dataOut.radarControllerHeaderObj.copy()

        self.processingHeaderObj.dtype = 0  # Voltage
        self.processingHeaderObj.blockSize = self.__getBlockSize()
        self.processingHeaderObj.profilesPerBlock = self.profilesPerBlock
        self.processingHeaderObj.dataBlocksPerFile = self.blocksPerFile
        # podria ser 1 o self.dataOut.processingHeaderObj.nWindows
        self.processingHeaderObj.nWindows = 1
        self.processingHeaderObj.nCohInt = self.dataOut.nCohInt
        # Cuando la data de origen es de tipo Voltage
        self.processingHeaderObj.nIncohInt = 1
        # Cuando la data de origen es de tipo Voltage
        self.processingHeaderObj.totalSpectra = 0

        if self.dataOut.code is not None:
            self.processingHeaderObj.code = self.dataOut.code
            self.processingHeaderObj.nCode = self.dataOut.nCode
            self.processingHeaderObj.nBaud = self.dataOut.nBaud

        if self.processingHeaderObj.nWindows != 0:
            self.processingHeaderObj.firstHeight = self.dataOut.heightList[0]
            self.processingHeaderObj.deltaHeight = self.dataOut.heightList[1] - \
                self.dataOut.heightList[0]
            self.processingHeaderObj.nHeights = self.dataOut.nHeights
            self.processingHeaderObj.samplesWin = self.dataOut.nHeights

        self.processingHeaderObj.processFlags = self.getProcessFlags()

        self.setBasicHeader()
        


'''
Created on Jul 2, 2014

@author: roj-idl71
'''
import numpy

from schainpy.model.io.jroIO_base import LOCALTIME, JRODataReader, JRODataWriter
from schainpy.model.proc.jroproc_base import ProcessingUnit, Operation, MPDecorator
from schainpy.model.data.jroheaderIO import PROCFLAG, BasicHeader, SystemHeader, RadarControllerHeader, ProcessingHeader
from schainpy.model.data.jrodata import Spectra
from schainpy.utils import log


class SpectraReader(JRODataReader, ProcessingUnit):
    """
    Esta clase permite leer datos de espectros desde archivos procesados (.pdata). La lectura
    de los datos siempre se realiza por bloques. Los datos leidos (array de 3 dimensiones)
    son almacenados en tres buffer's para el Self Spectra, el Cross Spectra y el DC Channel.

                        paresCanalesIguales    * alturas * perfiles  (Self Spectra)
                        paresCanalesDiferentes * alturas * perfiles  (Cross Spectra)
                        canales * alturas                            (DC Channels)

    Esta clase contiene instancias (objetos) de las clases BasicHeader, SystemHeader,
    RadarControllerHeader y Spectra. Los tres primeros se usan para almacenar informacion de la
    cabecera de datos (metadata), y el cuarto (Spectra) para obtener y almacenar un bloque de
    datos desde el "buffer" cada vez que se ejecute el metodo "getData".

    Example:
        dpath = "/home/myuser/data"

        startTime = datetime.datetime(2010,1,20,0,0,0,0,0,0)

        endTime = datetime.datetime(2010,1,21,23,59,59,0,0,0)

        readerObj = SpectraReader()

        readerObj.setup(dpath, startTime, endTime)

        while(True):

            readerObj.getData()

            print readerObj.data_spc

            print readerObj.data_cspc

            print readerObj.data_dc

            if readerObj.flagNoMoreFiles:
                break

    """

    def __init__(self):#, **kwargs):
        """
        Inicializador de la clase SpectraReader para la lectura de datos de espectros.

        Inputs:
            dataOut    :    Objeto de la clase Spectra. Este objeto sera utilizado para
                              almacenar un perfil de datos cada vez que se haga un requerimiento
                              (getData). El perfil sera obtenido a partir del buffer de datos,
                              si el buffer esta vacio se hara un nuevo proceso de lectura de un
                              bloque de datos.
                              Si este parametro no es pasado se creara uno internamente.

        Affected:
            self.dataOut

        Return      : None
        """

        ProcessingUnit.__init__(self)

        self.pts2read_SelfSpectra = 0
        self.pts2read_CrossSpectra = 0
        self.pts2read_DCchannels = 0        
        self.ext = ".pdata"
        self.optchar = "P"
        self.basicHeaderObj = BasicHeader(LOCALTIME)
        self.systemHeaderObj = SystemHeader()
        self.radarControllerHeaderObj = RadarControllerHeader()
        self.processingHeaderObj = ProcessingHeader()
        self.lastUTTime = 0
        self.maxTimeStep = 30
        self.dataOut = Spectra()
        self.profileIndex = 1
        self.nRdChannels = None
        self.nRdPairs = None
        self.rdPairList = []

    def createObjByDefault(self):

        dataObj = Spectra()

        return dataObj

    def __hasNotDataInBuffer(self):
        return 1


    def getBlockDimension(self):
        """
        Obtiene la cantidad de puntos a leer por cada bloque de datos

        Affected:
            self.nRdChannels
            self.nRdPairs
            self.pts2read_SelfSpectra
            self.pts2read_CrossSpectra
            self.pts2read_DCchannels
            self.blocksize
            self.dataOut.nChannels
            self.dataOut.nPairs

        Return:
            None
        """
        self.nRdChannels = 0
        self.nRdPairs = 0
        self.rdPairList = []

        for i in range(0, self.processingHeaderObj.totalSpectra*2, 2):
            if self.processingHeaderObj.spectraComb[i] == self.processingHeaderObj.spectraComb[i+1]:
                self.nRdChannels = self.nRdChannels + 1 #par de canales iguales
            else:
                self.nRdPairs = self.nRdPairs + 1 #par de canales diferentes
                self.rdPairList.append((self.processingHeaderObj.spectraComb[i], self.processingHeaderObj.spectraComb[i+1]))

        pts2read = self.processingHeaderObj.nHeights * self.processingHeaderObj.profilesPerBlock

        self.pts2read_SelfSpectra = int(self.nRdChannels * pts2read)
        self.blocksize = self.pts2read_SelfSpectra

        if self.processingHeaderObj.flag_cspc:
            self.pts2read_CrossSpectra = int(self.nRdPairs * pts2read)
            self.blocksize += self.pts2read_CrossSpectra

        if self.processingHeaderObj.flag_dc:
            self.pts2read_DCchannels = int(self.systemHeaderObj.nChannels * self.processingHeaderObj.nHeights)
            self.blocksize += self.pts2read_DCchannels

    def readBlock(self):
        """
        Lee el bloque de datos desde la posicion actual del puntero del archivo
        (self.fp) y actualiza todos los parametros relacionados al bloque de datos
        (metadata + data). La data leida es almacenada en el buffer y el contador del buffer
        es seteado a 0

        Return: None

        Variables afectadas:

            self.flagIsNewFile
            self.flagIsNewBlock
            self.nTotalBlocks
            self.data_spc
            self.data_cspc
            self.data_dc

        Exceptions:
            Si un bloque leido no es un bloque valido
        """
        
        fpointer = self.fp.tell()

        spc = numpy.fromfile( self.fp, self.dtype[0], self.pts2read_SelfSpectra )
        spc = spc.reshape( (self.nRdChannels, self.processingHeaderObj.nHeights, self.processingHeaderObj.profilesPerBlock) ) #transforma a un arreglo 3D

        if self.processingHeaderObj.flag_cspc:
            cspc = numpy.fromfile( self.fp, self.dtype, self.pts2read_CrossSpectra )
            cspc = cspc.reshape( (self.nRdPairs, self.processingHeaderObj.nHeights, self.processingHeaderObj.profilesPerBlock) ) #transforma a un arreglo 3D

        if self.processingHeaderObj.flag_dc:
            dc = numpy.fromfile( self.fp, self.dtype, self.pts2read_DCchannels ) #int(self.processingHeaderObj.nHeights*self.systemHeaderObj.nChannels) )
            dc = dc.reshape( (self.systemHeaderObj.nChannels, self.processingHeaderObj.nHeights) ) #transforma a un arreglo 2D

        if not self.processingHeaderObj.shif_fft:
            #desplaza a la derecha en el eje 2 determinadas posiciones
            shift = int(self.processingHeaderObj.profilesPerBlock/2)
            spc = numpy.roll( spc, shift , axis=2 )

            if self.processingHeaderObj.flag_cspc:
                #desplaza a la derecha en el eje 2 determinadas posiciones
                cspc = numpy.roll( cspc, shift, axis=2 )

        #Dimensions : nChannels, nProfiles, nSamples
        spc = numpy.transpose( spc, (0,2,1) )
        self.data_spc = spc

        if self.processingHeaderObj.flag_cspc:
            cspc = numpy.transpose( cspc, (0,2,1) )
            self.data_cspc = cspc['real'] + cspc['imag']*1j
        else:
            self.data_cspc = None

        if self.processingHeaderObj.flag_dc:
            self.data_dc = dc['real'] + dc['imag']*1j
        else:
            self.data_dc = None

        self.flagIsNewFile = 0
        self.flagIsNewBlock = 1

        self.nTotalBlocks += 1
        self.nReadBlocks += 1

        return 1

    def getFirstHeader(self):

        self.getBasicHeader()
        self.dataOut.systemHeaderObj = self.systemHeaderObj.copy()
        self.dataOut.radarControllerHeaderObj = self.radarControllerHeaderObj.copy()
        self.dataOut.dtype = self.dtype
        self.dataOut.pairsList = self.rdPairList
        self.dataOut.nProfiles = self.processingHeaderObj.profilesPerBlock
        self.dataOut.nFFTPoints = self.processingHeaderObj.profilesPerBlock
        self.dataOut.nCohInt = self.processingHeaderObj.nCohInt
        self.dataOut.nIncohInt = self.processingHeaderObj.nIncohInt
        xf = self.processingHeaderObj.firstHeight + self.processingHeaderObj.nHeights*self.processingHeaderObj.deltaHeight
        self.dataOut.heightList = numpy.arange(self.processingHeaderObj.firstHeight, xf, self.processingHeaderObj.deltaHeight)
        self.dataOut.channelList = list(range(self.systemHeaderObj.nChannels))
        self.dataOut.flagShiftFFT = True    #Data is always shifted
        self.dataOut.flagDecodeData = self.processingHeaderObj.flag_decode #asumo q la data no esta decodificada
        self.dataOut.flagDeflipData = self.processingHeaderObj.flag_deflip #asumo q la data esta sin flip

    def getData(self):
        """
        First method to execute before "RUN" is called.

        Copia el buffer de lectura a la clase "Spectra",
        con todos los parametros asociados a este (metadata). cuando no hay datos en el buffer de
        lectura es necesario hacer una nueva lectura de los bloques de datos usando "readNextBlock"

        Return:
            0    :    Si no hay mas archivos disponibles
            1    :    Si hizo una buena copia del buffer

        Affected:
            self.dataOut
            self.flagDiscontinuousBlock
            self.flagIsNewBlock
        """

        if self.flagNoMoreFiles:
            self.dataOut.flagNoData = True
            return 0

        self.flagDiscontinuousBlock = 0
        self.flagIsNewBlock = 0

        if self.__hasNotDataInBuffer():

            if not( self.readNextBlock() ):
                self.dataOut.flagNoData = True
                return 0

        #data es un numpy array de 3 dmensiones (perfiles, alturas y canales)

        if self.data_spc is None:
            self.dataOut.flagNoData = True
            return 0

        self.getBasicHeader()
        self.getFirstHeader()
        self.dataOut.data_spc = self.data_spc
        self.dataOut.data_cspc = self.data_cspc
        self.dataOut.data_dc = self.data_dc
        self.dataOut.flagNoData = False
        self.dataOut.realtime = self.online

        return self.dataOut.data_spc


@MPDecorator
class SpectraWriter(JRODataWriter, Operation):

    """
    Esta clase permite escribir datos de espectros a archivos procesados (.pdata). La escritura
    de los datos siempre se realiza por bloques.
    """

    def __init__(self):
        """
        Inicializador de la clase SpectraWriter para la escritura de datos de espectros.

        Affected:
            self.dataOut
            self.basicHeaderObj
            self.systemHeaderObj
            self.radarControllerHeaderObj
            self.processingHeaderObj

        Return: None
        """

        Operation.__init__(self)

        self.ext = ".pdata"
        self.optchar = "P"
        self.shape_spc_Buffer = None
        self.shape_cspc_Buffer = None
        self.shape_dc_Buffer = None
        self.data_spc = None
        self.data_cspc = None
        self.data_dc = None
        self.setFile = None
        self.noMoreFiles = 0
        self.basicHeaderObj = BasicHeader(LOCALTIME)
        self.systemHeaderObj = SystemHeader()
        self.radarControllerHeaderObj = RadarControllerHeader()
        self.processingHeaderObj = ProcessingHeader()

    def hasAllDataInBuffer(self):
        return 1


    def setBlockDimension(self):
        """
        Obtiene las formas dimensionales del los subbloques de datos que componen un bloque

        Affected:
            self.shape_spc_Buffer
            self.shape_cspc_Buffer
            self.shape_dc_Buffer

        Return: None
        """
        self.shape_spc_Buffer = (self.dataOut.nChannels,
                                 self.processingHeaderObj.nHeights,
                                 self.processingHeaderObj.profilesPerBlock)

        self.shape_cspc_Buffer = (self.dataOut.nPairs,
                                  self.processingHeaderObj.nHeights,
                                  self.processingHeaderObj.profilesPerBlock)

        self.shape_dc_Buffer = (self.dataOut.nChannels,
                                self.processingHeaderObj.nHeights)


    def writeBlock(self):
        """processingHeaderObj
        Escribe el buffer en el file designado

        Affected:
            self.data_spc
            self.data_cspc
            self.data_dc
            self.flagIsNewFile
            self.flagIsNewBlock
            self.nTotalBlocks
            self.nWriteBlocks

        Return: None
        """

        spc = numpy.transpose( self.data_spc, (0,2,1) )
        if not self.processingHeaderObj.shif_fft:
            spc = numpy.roll( spc, int(self.processingHeaderObj.profilesPerBlock/2), axis=2 ) #desplaza a la derecha en el eje 2 determinadas posiciones
        data = spc.reshape((-1))
        data = data.astype(self.dtype[0])
        data.tofile(self.fp)

        if self.data_cspc is not None:
            
            cspc = numpy.transpose( self.data_cspc, (0,2,1) )
            data = numpy.zeros( numpy.shape(cspc), self.dtype )
            #print 'data.shape', self.shape_cspc_Buffer
            if not self.processingHeaderObj.shif_fft:
                cspc = numpy.roll( cspc, int(self.processingHeaderObj.profilesPerBlock/2), axis=2 ) #desplaza a la derecha en el eje 2 determinadas posiciones
            data['real'] = cspc.real
            data['imag'] = cspc.imag
            data = data.reshape((-1))
            data.tofile(self.fp)

        if self.data_dc is not None:
            
            dc = self.data_dc
            data = numpy.zeros( numpy.shape(dc), self.dtype )
            data['real'] = dc.real
            data['imag'] = dc.imag
            data = data.reshape((-1))
            data.tofile(self.fp)

#         self.data_spc.fill(0)
#
#         if self.data_dc is not None:
#             self.data_dc.fill(0)
#
#         if self.data_cspc is not None:
#             self.data_cspc.fill(0)

        self.flagIsNewFile = 0
        self.flagIsNewBlock = 1
        self.nTotalBlocks += 1
        self.nWriteBlocks += 1
        self.blockIndex += 1

#         print "[Writing] Block = %d04" %self.blockIndex

    def putData(self):
        """
        Setea un bloque de datos y luego los escribe en un file

        Affected:
            self.data_spc
            self.data_cspc
            self.data_dc

        Return:
            0    :    Si no hay data o no hay mas files que puedan escribirse
            1    :    Si se escribio la data de un bloque en un file
        """

        if self.dataOut.flagNoData:
            return 0

        self.flagIsNewBlock = 0

        if self.dataOut.flagDiscontinuousBlock:
            self.data_spc.fill(0)
            if self.dataOut.data_cspc is not None:
                self.data_cspc.fill(0)
            if self.dataOut.data_dc is not None:
                self.data_dc.fill(0)
            self.setNextFile()

        if self.flagIsNewFile == 0:
            self.setBasicHeader()

        self.data_spc = self.dataOut.data_spc.copy()

        if self.dataOut.data_cspc is not None:
            self.data_cspc = self.dataOut.data_cspc.copy()

        if self.dataOut.data_dc is not None:
            self.data_dc = self.dataOut.data_dc.copy()

        # #self.processingHeaderObj.dataBlocksPerFile)
        if self.hasAllDataInBuffer():
#            self.setFirstHeader()
            self.writeNextBlock()

    def __getBlockSize(self):
        '''
        Este metodos determina el cantidad de bytes para un bloque de datos de tipo Spectra
        '''

        dtype_width = self.getDtypeWidth()

        pts2write = self.dataOut.nHeights * self.dataOut.nFFTPoints

        pts2write_SelfSpectra = int(self.dataOut.nChannels * pts2write)
        blocksize = (pts2write_SelfSpectra*dtype_width)

        if self.dataOut.data_cspc is not None:
            pts2write_CrossSpectra = int(self.dataOut.nPairs * pts2write)
            blocksize += (pts2write_CrossSpectra*dtype_width*2)

        if self.dataOut.data_dc is not None:
            pts2write_DCchannels = int(self.dataOut.nChannels * self.dataOut.nHeights)
            blocksize += (pts2write_DCchannels*dtype_width*2)

#         blocksize = blocksize #* datatypeValue * 2 #CORREGIR ESTO

        return blocksize

    def setFirstHeader(self):

        """
        Obtiene una copia del First Header

        Affected:
            self.systemHeaderObj
            self.radarControllerHeaderObj
            self.dtype

        Return:
            None
        """

        self.systemHeaderObj = self.dataOut.systemHeaderObj.copy()
        self.systemHeaderObj.nChannels = self.dataOut.nChannels
        self.radarControllerHeaderObj = self.dataOut.radarControllerHeaderObj.copy()

        self.processingHeaderObj.dtype = 1 # Spectra
        self.processingHeaderObj.blockSize = self.__getBlockSize()
        self.processingHeaderObj.profilesPerBlock = self.dataOut.nFFTPoints
        self.processingHeaderObj.dataBlocksPerFile = self.blocksPerFile
        self.processingHeaderObj.nWindows = 1 #podria ser 1 o self.dataOut.processingHeaderObj.nWindows
        self.processingHeaderObj.nCohInt = self.dataOut.nCohInt# Se requiere para determinar el valor de timeInterval
        self.processingHeaderObj.nIncohInt = self.dataOut.nIncohInt
        self.processingHeaderObj.totalSpectra = self.dataOut.nPairs + self.dataOut.nChannels
        self.processingHeaderObj.shif_fft = self.dataOut.flagShiftFFT

        if self.processingHeaderObj.totalSpectra > 0:
            channelList = []
            for channel in range(self.dataOut.nChannels):
                channelList.append(channel)
                channelList.append(channel)

            pairsList = []
            if self.dataOut.nPairs > 0:
                for pair in self.dataOut.pairsList:
                    pairsList.append(pair[0])
                    pairsList.append(pair[1])

            spectraComb = channelList + pairsList
            spectraComb = numpy.array(spectraComb, dtype="u1")
            self.processingHeaderObj.spectraComb = spectraComb

        if self.dataOut.code is not None:
            self.processingHeaderObj.code = self.dataOut.code
            self.processingHeaderObj.nCode = self.dataOut.nCode
            self.processingHeaderObj.nBaud = self.dataOut.nBaud

        if self.processingHeaderObj.nWindows != 0:
            self.processingHeaderObj.firstHeight = self.dataOut.heightList[0]
            self.processingHeaderObj.deltaHeight = self.dataOut.heightList[1] - self.dataOut.heightList[0]
            self.processingHeaderObj.nHeights = self.dataOut.nHeights
            self.processingHeaderObj.samplesWin = self.dataOut.nHeights

        self.processingHeaderObj.processFlags = self.getProcessFlags()

        self.setBasicHeader()


'''

$Author: murco $
$Id: JROHeaderIO.py 151 2012-10-31 19:00:51Z murco $
'''
import sys
import numpy
import copy
import datetime


SPEED_OF_LIGHT = 299792458
SPEED_OF_LIGHT = 3e8

FILE_STRUCTURE = numpy.dtype([          #HEADER 48bytes
                          ('FileMgcNumber','<u4'),      #0x23020100
                          ('nFDTdataRecors','<u4'),     #No Of FDT data records in this file (0 or more)
                          ('RadarUnitId','<u4'),
                          ('SiteName','<s32'),          #Null terminated
                          ])

RECORD_STRUCTURE = numpy.dtype([         #RECORD HEADER 180+20N bytes
                            ('RecMgcNumber','<u4'),     #0x23030001
                            ('RecCounter','<u4'),       #Record counter(0,1, ...)
                            ('Off2StartNxtRec','<u4'),  #Offset to start of next record form start of this record
                            ('Off2StartData','<u4'),    #Offset to start of data from start of this record     
                            ('EpTimeStamp','<i4'),      #Epoch time stamp of start of acquisition (seconds)   
                            ('msCompTimeStamp','<u4'),  #Millisecond component of time stamp (0,...,999) 
                            ('ExpTagName','<s32'),      #Experiment tag name (null terminated)
                            ('ExpComment','<s32'),      #Experiment comment (null terminated)
                            ('SiteLatDegrees','<f4'),   #Site latitude (from GPS) in degrees (positive implies North)
                            ('SiteLongDegrees','<f4'),  #Site longitude (from GPS) in degrees (positive implies East)
                            ('RTCgpsStatus','<u4'),     #RTC GPS engine status (0=SEEK, 1=LOCK, 2=NOT FITTED, 3=UNAVAILABLE)
                            ('TransmitFrec','<u4'),     #Transmit frequency (Hz)
                            ('ReceiveFrec','<u4'),      #Receive frequency
                            ('FirstOsciFrec','<u4'),    #First local oscillator frequency (Hz) 
                            ('Polarisation','<u4'),     #(0="O", 1="E", 2="linear 1", 3="linear2")
                            ('ReceiverFiltSett','<u4'), #Receiver filter settings (0,1,2,3)
                            ('nModesInUse','<u4'),      #Number of modes in use (1 or 2)
                            ('DualModeIndex','<u4'),    #Dual Mode index number for these data (0 or 1)
                            ('DualModeRange','<u4'),    #Dual Mode range correction for these data (m)
                            ('nDigChannels','<u4'),     #Number of digital channels acquired (2*N)
                            ('SampResolution','<u4'),   #Sampling resolution (meters)
                            ('nRangeGatesSamp','<u4'),  #Number of range gates sampled
                            ('StartRangeSamp','<u4'),   #Start range of sampling (meters)
                            ('PRFhz','<u4'),            #PRF (Hz)
                            ('Integrations','<u4'),     #Integrations
                            ('nDataPointsTrsf','<u4'),  #Number of data points transformed
                            ('nReceiveBeams','<u4'),    #Number of receive beams stored in file (1 or N)
                            ('nSpectAverages','<u4'),   #Number of spectral averages
                            ('FFTwindowingInd','<u4'),  #FFT windowing index (0 = no window)
                            ('BeamAngleAzim','<f4'),    #Beam steer angle (azimuth) in degrees (clockwise from true North)
                            ('BeamAngleZen','<f4'),     #Beam steer angle (zenith) in degrees (0=> vertical)
                            ('AntennaCoord','<f24'),     #Antenna coordinates (Range(meters), Bearing(degrees)) - N pairs
                            ('RecPhaseCalibr','<f12'),   #Receiver phase calibration (degrees) - N values
                            ('RecAmpCalibr','<f12'),     #Receiver amplitude calibration (ratio relative to receiver one) - N values
                            ('ReceiverGaindB','<u12'),   #Receiver gains in dB - N values
                            ])


class Header(object):
    
    def __init__(self):
        raise NotImplementedError
    
   
    def read(self):
        
        raise NotImplementedError
    
    def write(self):
        
        raise NotImplementedError
    
    def printInfo(self):
        
        message = "#"*50 + "\n"
        message += self.__class__.__name__.upper() + "\n"
        message += "#"*50 + "\n"
        
        keyList = list(self.__dict__.keys())
        keyList.sort()
        
        for key in keyList:
            message += "%s = %s" %(key, self.__dict__[key]) + "\n"
        
        if "size" not in keyList:
            attr = getattr(self, "size")
            
            if attr:
                message += "%s = %s" %("size", attr) + "\n"
        
        print(message)

class FileHeader(Header):
    
    FileMgcNumber= None
    nFDTdataRecors=None     #No Of FDT data records in this file (0 or more)
    RadarUnitId= None
    SiteName= None
 
    #__LOCALTIME = None
        
    def __init__(self, useLocalTime=True):
        
        self.FileMgcNumber= 0     #0x23020100
        self.nFDTdataRecors=0     #No Of FDT data records in this file (0 or more)
        self.RadarUnitId= 0
        self.SiteName= ""
        self.size = 48
 
        #self.useLocalTime = useLocalTime
    
    def read(self, fp):
        
        try:
            header = numpy.fromfile(fp, FILE_STRUCTURE,1)
            '''      numpy.fromfile(file, dtype, count, sep='')
            file : file or str
            Open file object or filename.

            dtype : data-type
            Data type of the returned array. For binary files, it is used to determine 
            the size and byte-order of the items in the file.

            count : int
            Number of items to read. -1 means all items (i.e., the complete file).

            sep : str
            Separator between items if file is a text file. Empty (“”) separator means 
            the file should be treated as binary. Spaces (” ”) in the separator match zero 
            or more whitespace characters. A separator consisting only of spaces must match 
            at least one whitespace.

            '''
            
        except Exception as e:
            print("FileHeader: ")
            print(eBasicHeader)
            return 0
        
        self.FileMgcNumber= byte(header['FileMgcNumber'][0])
        self.nFDTdataRecors=int(header['nFDTdataRecors'][0])     #No Of FDT data records in this file (0 or more)
        self.RadarUnitId= int(header['RadarUnitId'][0])
        self.SiteName= char(header['SiteName'][0])
        

        if self.size <48:
            return 0
            
        return 1
    
    def write(self, fp):
        
        headerTuple = (self.FileMgcNumber,
                       self.nFDTdataRecors,
                       self.RadarUnitId,
                       self.SiteName,
                       self.size)
             
                       
        header = numpy.array(headerTuple, FILE_STRUCTURE)
        #        numpy.array(object, dtype=None, copy=True, order=None, subok=False, ndmin=0)
        header.tofile(fp)
        ''' ndarray.tofile(fid, sep, format)    Write array to a file as text or binary (default).

        fid : file or str
        An open file object, or a string containing a filename.

        sep : str
        Separator between array items for text output. If “” (empty), a binary file is written, 
        equivalent to file.write(a.tobytes()).

        format : str
        Format string for text file output. Each entry in the array is formatted to text by 
        first converting it to the closest Python type, and then using “format” % item.

        '''
        
        return 1
    

class RecordHeader(Header):
    
    RecMgcNumber=None     #0x23030001
    RecCounter= None      
    Off2StartNxtRec= None      
    EpTimeStamp= None        
    msCompTimeStamp= None  
    ExpTagName= None      
    ExpComment=None      
    SiteLatDegrees=None  
    SiteLongDegrees= None 
    RTCgpsStatus= None    
    TransmitFrec= None    
    ReceiveFrec= None     
    FirstOsciFrec= None    
    Polarisation= None    
    ReceiverFiltSett= None
    nModesInUse= None     
    DualModeIndex= None   
    DualModeRange= None   
    nDigChannels= None
    SampResolution= None
    nRangeGatesSamp= None
    StartRangeSamp= None
    PRFhz= None
    Integrations= None
    nDataPointsTrsf= None
    nReceiveBeams= None
    nSpectAverages= None
    FFTwindowingInd= None
    BeamAngleAzim= None
    BeamAngleZen= None
    AntennaCoord= None
    RecPhaseCalibr= None
    RecAmpCalibr= None
    ReceiverGaindB= None
    
    '''size = None
    nSamples = None
    nProfiles = None
    nChannels = None
    adcResolution = None
    pciDioBusWidth = None'''
        
    def __init__(self,      RecMgcNumber=None,    RecCounter= 0,       Off2StartNxtRec= 0,      
                            EpTimeStamp= 0,    msCompTimeStamp= 0,  ExpTagName= None,      
                            ExpComment=None,      SiteLatDegrees=0,    SiteLongDegrees= 0, 
                            RTCgpsStatus= 0,   TransmitFrec= 0,     ReceiveFrec= 0,     
                            FirstOsciFrec= 0,  Polarisation= 0,     ReceiverFiltSett= 0,
                            nModesInUse= 0,    DualModeIndex= 0,    DualModeRange= 0,   
                            nDigChannels= 0,   SampResolution= 0,   nRangeGatesSamp= 0,
                            StartRangeSamp= 0, PRFhz= 0,            Integrations= 0,
                            nDataPointsTrsf= 0,  nReceiveBeams= 0,  nSpectAverages= 0,
                            FFTwindowingInd= 0,  BeamAngleAzim= 0,  BeamAngleZen= 0,
                            AntennaCoord= 0,   RecPhaseCalibr= 0,   RecAmpCalibr= 0,
                            ReceiverGaindB= 0):
        
        self.RecMgcNumber = RecMgcNumber     #0x23030001
        self.RecCounter = RecCounter      
        self.Off2StartNxtRec = Off2StartNxtRec       
        self.EpTimeStamp = EpTimeStamp         
        self.msCompTimeStamp = msCompTimeStamp   
        self.ExpTagName = ExpTagName       
        self.ExpComment = ExpComment      
        self.SiteLatDegrees = SiteLatDegrees  
        self.SiteLongDegrees = SiteLongDegrees  
        self.RTCgpsStatus = RTCgpsStatus     
        self.TransmitFrec = TransmitFrec     
        self.ReceiveFrec = ReceiveFrec      
        self.FirstOsciFrec = FirstOsciFrec     
        self.Polarisation = Polarisation     
        self.ReceiverFiltSett = ReceiverFiltSett 
        self.nModesInUse = nModesInUse      
        self.DualModeIndex = DualModeIndex    
        self.DualModeRange = DualModeRange    
        self.nDigChannels = nDigChannels
        self.SampResolution = SampResolution 
        self.nRangeGatesSamp = nRangeGatesSamp 
        self.StartRangeSamp = StartRangeSamp 
        self.PRFhz = PRFhz 
        self.Integrations = Integrations 
        self.nDataPointsTrsf = nDataPointsTrsf 
        self.nReceiveBeams = nReceiveBeams 
        self.nSpectAverages = nSpectAverages 
        self.FFTwindowingInd = FFTwindowingInd 
        self.BeamAngleAzim = BeamAngleAzim 
        self.BeamAngleZen = BeamAngleZen 
        self.AntennaCoord = AntennaCoord 
        self.RecPhaseCalibr = RecPhaseCalibr 
        self.RecAmpCalibr = RecAmpCalibr 
        self.ReceiverGaindB = ReceiverGaindB 
        
        
    def read(self, fp):
        
        startFp = fp.tell() #The method tell() returns the current position of the file read/write pointer within the file.
        
        try:
            header = numpy.fromfile(fp,RECORD_STRUCTURE,1)
        except Exception as e:
            print("System Header: " + e)
            return 0
        
        self.RecMgcNumber = header['RecMgcNumber'][0]     #0x23030001
        self.RecCounter = header['RecCounter'][0]      
        self.Off2StartNxtRec = header['Off2StartNxtRec'][0]       
        self.EpTimeStamp = header['EpTimeStamp'][0]         
        self.msCompTimeStamp = header['msCompTimeStamp'][0]   
        self.ExpTagName = header['ExpTagName'][0]       
        self.ExpComment = header['ExpComment'][0]      
        self.SiteLatDegrees = header['SiteLatDegrees'][0]  
        self.SiteLongDegrees = header['SiteLongDegrees'][0]  
        self.RTCgpsStatus = header['RTCgpsStatus'][0]     
        self.TransmitFrec = header['TransmitFrec'][0]     
        self.ReceiveFrec = header['ReceiveFrec'][0]      
        self.FirstOsciFrec = header['FirstOsciFrec'][0]     
        self.Polarisation = header['Polarisation'][0]     
        self.ReceiverFiltSett = header['ReceiverFiltSett'][0] 
        self.nModesInUse = header['nModesInUse'][0]      
        self.DualModeIndex = header['DualModeIndex'][0]    
        self.DualModeRange = header['DualModeRange'][0]    
        self.nDigChannels = header['nDigChannels'][0]
        self.SampResolution = header['SampResolution'][0] 
        self.nRangeGatesSamp = header['nRangeGatesSamp'][0] 
        self.StartRangeSamp = header['StartRangeSamp'][0] 
        self.PRFhz = header['PRFhz'][0] 
        self.Integrations = header['Integrations'][0] 
        self.nDataPointsTrsf = header['nDataPointsTrsf'][0] 
        self.nReceiveBeams = header['nReceiveBeams'][0] 
        self.nSpectAverages = header['nSpectAverages'][0] 
        self.FFTwindowingInd = header['FFTwindowingInd'][0] 
        self.BeamAngleAzim = header['BeamAngleAzim'][0] 
        self.BeamAngleZen = header['BeamAngleZen'][0] 
        self.AntennaCoord = header['AntennaCoord'][0] 
        self.RecPhaseCalibr = header['RecPhaseCalibr'][0] 
        self.RecAmpCalibr = header['RecAmpCalibr'][0] 
        self.ReceiverGaindB = header['ReceiverGaindB'][0]
        
        Self.size = 180+20*3
        
        endFp = self.size + startFp
        
        if fp.tell() > endFp:
            sys.stderr.write("Warning %s: Size value read from System Header is lower than it has to be\n" %fp.name)
            return 0
            
        if fp.tell() < endFp:
            sys.stderr.write("Warning %s: Size value read from System Header size is greater than it has to be\n" %fp.name)
            return 0
        
        return 1
    
    def write(self, fp):
        
        headerTuple = (self.RecMgcNumber,
                       self.RecCounter,      
                       self.Off2StartNxtRec,       
                       self.EpTimeStamp,         
                       self.msCompTimeStamp,   
                       self.ExpTagName,       
                       self.ExpComment,      
                       self.SiteLatDegrees,  
                       self.SiteLongDegrees,  
                       self.RTCgpsStatus,     
                       self.TransmitFrec,     
                       self.ReceiveFrec,      
                       self.FirstOsciFrec,     
                       self.Polarisation,     
                       self.ReceiverFiltSett, 
                       self.nModesInUse,      
                       self.DualModeIndex,    
                       self.DualModeRange,    
                       self.nDigChannels,
                       self.SampResolution, 
                       self.nRangeGatesSamp, 
                       self.StartRangeSamp, 
                       self.PRFhz, 
                       self.Integrations, 
                       self.nDataPointsTrsf, 
                       self.nReceiveBeams, 
                       self.nSpectAverages, 
                       self.FFTwindowingInd, 
                       self.BeamAngleAzim, 
                       self.BeamAngleZen, 
                       self.AntennaCoord, 
                       self.RecPhaseCalibr, 
                       self.RecAmpCalibr, 
                       self.ReceiverGaindB) 
                       
#                        self.size,self.nSamples,
#                        self.nProfiles,
#                        self.nChannels,
#                        self.adcResolution,
#                        self.pciDioBusWidth
                       
        header = numpy.array(headerTuple,RECORD_STRUCTURE)
        header.tofile(fp)
        
        return 1

    
def get_dtype_index(numpy_dtype):
    
    index = None
    
    for i in range(len(NUMPY_DTYPE_LIST)):
        if numpy_dtype == NUMPY_DTYPE_LIST[i]:
            index = i
            break
    
    return index

def get_numpy_dtype(index):
    
    #dtype4 = numpy.dtype([('real','<f4'),('imag','<f4')])
    
    return NUMPY_DTYPE_LIST[index]


def get_dtype_width(index):
    
    return DTYPE_WIDTH[index]


import os
import sys
import glob
import fnmatch
import datetime
import time
import re
import h5py
import numpy

import pylab as plb
from scipy.optimize import curve_fit
from scipy import asarray as ar, exp

SPEED_OF_LIGHT = 299792458
SPEED_OF_LIGHT = 3e8

from .utils import folder_in_range

import schainpy.admin
from schainpy.model.data.jrodata import Spectra
from schainpy.model.proc.jroproc_base import ProcessingUnit, Operation, MPDecorator
from schainpy.utils import log
from schainpy.model.io.jroIO_base import JRODataReader

def pol2cart(rho, phi):
    x = rho * numpy.cos(phi)
    y = rho * numpy.sin(phi)
    return(x, y)

FILE_STRUCTURE = numpy.dtype([  # HEADER 48bytes
    ('FileMgcNumber', '<u4'),  # 0x23020100
    ('nFDTdataRecors', '<u4'),
    ('OffsetStartHeader', '<u4'),
    ('RadarUnitId', '<u4'),
    ('SiteName', 'S32'),  # Null terminated
])


class FileHeaderBLTR():

    def __init__(self, fo):

        self.fo = fo
        self.size = 48
        self.read()

    def read(self):

        header = numpy.fromfile(self.fo, FILE_STRUCTURE, 1)
        self.FileMgcNumber = hex(header['FileMgcNumber'][0])
        self.nFDTdataRecors = int(header['nFDTdataRecors'][0])
        self.RadarUnitId = int(header['RadarUnitId'][0])
        self.OffsetStartHeader = int(header['OffsetStartHeader'][0])
        self.SiteName = header['SiteName'][0]

    def write(self, fp):

        headerTuple = (self.FileMgcNumber,
                       self.nFDTdataRecors,
                       self.RadarUnitId,
                       self.SiteName,
                       self.size)

        header = numpy.array(headerTuple, FILE_STRUCTURE)
        header.tofile(fp)
        ''' ndarray.tofile(fid, sep, format)    Write array to a file as text or binary (default).

        fid : file or str
        An open file object, or a string containing a filename.

        sep : str
        Separator between array items for text output. If "" (empty), a binary file is written, 
        equivalent to file.write(a.tobytes()).

        format : str
        Format string for text file output. Each entry in the array is formatted to text by 
        first converting it to the closest Python type, and then using "format" % item.

        '''

        return 1


RECORD_STRUCTURE = numpy.dtype([  # RECORD HEADER 180+20N bytes
    ('RecMgcNumber', '<u4'),  # 0x23030001
    ('RecCounter', '<u4'),  # Record counter(0,1, ...)
    # Offset to start of next record form start of this record
                            ('Off2StartNxtRec', '<u4'),
                            # Offset to start of data from start of this record
                            ('Off2StartData', '<u4'),
                            # Epoch time stamp of start of acquisition (seconds)
                            ('nUtime', '<i4'),
                            # Millisecond component of time stamp (0,...,999)
                            ('nMilisec', '<u4'),
                            # Experiment tag name (null terminated)
                            ('ExpTagName', 'S32'),
                            # Experiment comment (null terminated)
                            ('ExpComment', 'S32'),
                            # Site latitude (from GPS) in degrees (positive implies North)
                            ('SiteLatDegrees', '<f4'),
                            # Site longitude (from GPS) in degrees (positive implies East)
                            ('SiteLongDegrees', '<f4'),
                            # RTC GPS engine status (0=SEEK, 1=LOCK, 2=NOT FITTED, 3=UNAVAILABLE)
                            ('RTCgpsStatus', '<u4'),
                            ('TransmitFrec', '<u4'),  # Transmit frequency (Hz)
                            ('ReceiveFrec', '<u4'),  # Receive frequency
                            # First local oscillator frequency (Hz)
                            ('FirstOsciFrec', '<u4'),
                            # (0="O", 1="E", 2="linear 1", 3="linear2")
                            ('Polarisation', '<u4'),
                            # Receiver filter settings (0,1,2,3)
                            ('ReceiverFiltSett', '<u4'),
                            # Number of modes in use (1 or 2)
                            ('nModesInUse', '<u4'),
                            # Dual Mode index number for these data (0 or 1)
                            ('DualModeIndex', '<u4'),
                            # Dual Mode range correction for these data (m)
                            ('DualModeRange', '<u4'),
                            # Number of digital channels acquired (2*N)
                            ('nDigChannels', '<u4'),
                            # Sampling resolution (meters)
                            ('SampResolution', '<u4'),
                            # Number of range gates sampled
                            ('nHeights', '<u4'),
                            # Start range of sampling (meters)
                            ('StartRangeSamp', '<u4'),
                            ('PRFhz', '<u4'),  # PRF (Hz)
                            ('nCohInt', '<u4'),  # Integrations
                            # Number of data points transformed
                            ('nProfiles', '<u4'),
                            # Number of receive beams stored in file (1 or N)
                            ('nChannels', '<u4'),
                            ('nIncohInt', '<u4'),  # Number of spectral averages
                            # FFT windowing index (0 = no window)
                            ('FFTwindowingInd', '<u4'),
                            # Beam steer angle (azimuth) in degrees (clockwise from true North)
                            ('BeamAngleAzim', '<f4'),
                            # Beam steer angle (zenith) in degrees (0=> vertical)
                            ('BeamAngleZen', '<f4'),
                            # Antenna coordinates (Range(meters), Bearing(degrees)) - N pairs
                            ('AntennaCoord0', '<f4'),
                            # Antenna coordinates (Range(meters), Bearing(degrees)) - N pairs
                            ('AntennaAngl0', '<f4'),
                            # Antenna coordinates (Range(meters), Bearing(degrees)) - N pairs
                            ('AntennaCoord1', '<f4'),
                            # Antenna coordinates (Range(meters), Bearing(degrees)) - N pairs
                            ('AntennaAngl1', '<f4'),
                            # Antenna coordinates (Range(meters), Bearing(degrees)) - N pairs
                            ('AntennaCoord2', '<f4'),
                            # Antenna coordinates (Range(meters), Bearing(degrees)) - N pairs
                            ('AntennaAngl2', '<f4'),
                            # Receiver phase calibration (degrees) - N values
                            ('RecPhaseCalibr0', '<f4'),
                            # Receiver phase calibration (degrees) - N values
                            ('RecPhaseCalibr1', '<f4'),
                            # Receiver phase calibration (degrees) - N values
                            ('RecPhaseCalibr2', '<f4'),
                            # Receiver amplitude calibration (ratio relative to receiver one) - N values
                            ('RecAmpCalibr0', '<f4'),
                            # Receiver amplitude calibration (ratio relative to receiver one) - N values
                            ('RecAmpCalibr1', '<f4'),
                            # Receiver amplitude calibration (ratio relative to receiver one) - N values
                            ('RecAmpCalibr2', '<f4'),
                            # Receiver gains in dB - N values
                            ('ReceiverGaindB0', '<i4'),
                            # Receiver gains in dB - N values
                            ('ReceiverGaindB1', '<i4'),
                            # Receiver gains in dB - N values
                            ('ReceiverGaindB2', '<i4'),
])


class RecordHeaderBLTR():

    def __init__(self, fo):

        self.fo = fo
        self.OffsetStartHeader = 48
        self.Off2StartNxtRec = 811248

    def read(self, block):
        OffRHeader = self.OffsetStartHeader + block * self.Off2StartNxtRec
        self.fo.seek(OffRHeader, os.SEEK_SET)
        header = numpy.fromfile(self.fo, RECORD_STRUCTURE, 1)
        self.RecMgcNumber = hex(header['RecMgcNumber'][0])  # 0x23030001
        self.RecCounter = int(header['RecCounter'][0])
        self.Off2StartNxtRec = int(header['Off2StartNxtRec'][0])
        self.Off2StartData = int(header['Off2StartData'][0])
        self.nUtime = header['nUtime'][0]
        self.nMilisec = header['nMilisec'][0]
        self.ExpTagName = '' # str(header['ExpTagName'][0])
        self.ExpComment = '' # str(header['ExpComment'][0])
        self.SiteLatDegrees = header['SiteLatDegrees'][0]
        self.SiteLongDegrees = header['SiteLongDegrees'][0]
        self.RTCgpsStatus = header['RTCgpsStatus'][0]
        self.TransmitFrec = header['TransmitFrec'][0]
        self.ReceiveFrec = header['ReceiveFrec'][0]
        self.FirstOsciFrec = header['FirstOsciFrec'][0]
        self.Polarisation = header['Polarisation'][0]
        self.ReceiverFiltSett = header['ReceiverFiltSett'][0]
        self.nModesInUse = header['nModesInUse'][0]
        self.DualModeIndex = header['DualModeIndex'][0]
        self.DualModeRange = header['DualModeRange'][0]
        self.nDigChannels = header['nDigChannels'][0]
        self.SampResolution = header['SampResolution'][0]
        self.nHeights = header['nHeights'][0]
        self.StartRangeSamp = header['StartRangeSamp'][0]
        self.PRFhz = header['PRFhz'][0]
        self.nCohInt = header['nCohInt'][0]
        self.nProfiles = header['nProfiles'][0]
        self.nChannels = header['nChannels'][0]
        self.nIncohInt = header['nIncohInt'][0]
        self.FFTwindowingInd = header['FFTwindowingInd'][0]
        self.BeamAngleAzim = header['BeamAngleAzim'][0]
        self.BeamAngleZen = header['BeamAngleZen'][0]
        self.AntennaCoord0 = header['AntennaCoord0'][0]
        self.AntennaAngl0 = header['AntennaAngl0'][0]
        self.AntennaCoord1 = header['AntennaCoord1'][0]
        self.AntennaAngl1 = header['AntennaAngl1'][0]
        self.AntennaCoord2 = header['AntennaCoord2'][0]
        self.AntennaAngl2 = header['AntennaAngl2'][0]
        self.RecPhaseCalibr0 = header['RecPhaseCalibr0'][0]
        self.RecPhaseCalibr1 = header['RecPhaseCalibr1'][0]
        self.RecPhaseCalibr2 = header['RecPhaseCalibr2'][0]
        self.RecAmpCalibr0 = header['RecAmpCalibr0'][0]
        self.RecAmpCalibr1 = header['RecAmpCalibr1'][0]
        self.RecAmpCalibr2 = header['RecAmpCalibr2'][0]
        self.ReceiverGaindB0 = header['ReceiverGaindB0'][0]
        self.ReceiverGaindB1 = header['ReceiverGaindB1'][0]
        self.ReceiverGaindB2 = header['ReceiverGaindB2'][0]
        self.ipp = 0.5 * (SPEED_OF_LIGHT / self.PRFhz)
        self.RHsize = 180 + 20 * self.nChannels
        self.Datasize = self.nProfiles * self.nChannels * self.nHeights * 2 * 4
        endFp = self.OffsetStartHeader + self.RecCounter * self.Off2StartNxtRec

        
        if OffRHeader > endFp:
            sys.stderr.write(
                "Warning %s: Size value read from System Header is lower than it has to be\n" % fp)
            return 0

        if OffRHeader < endFp:
            sys.stderr.write(
                "Warning %s: Size value read from System Header size is greater than it has to be\n" % fp)
            return 0

        return 1


class BLTRSpectraReader (ProcessingUnit):

    def __init__(self):

        ProcessingUnit.__init__(self)

        self.ext = ".fdt"
        self.optchar = "P"
        self.fpFile = None
        self.fp = None
        self.BlockCounter = 0
        self.fileSizeByHeader = None
        self.filenameList = []
        self.fileSelector = 0
        self.Off2StartNxtRec = 0
        self.RecCounter = 0
        self.flagNoMoreFiles = 0
        self.data_spc = None
        self.data_cspc = None
        self.path = None
        self.OffsetStartHeader = 0
        self.Off2StartData = 0
        self.ipp = 0
        self.nFDTdataRecors = 0
        self.blocksize = 0
        self.dataOut = Spectra()
        self.dataOut.flagNoData = False

    def search_files(self):
        ''' 
        Function that indicates the number of .fdt files that exist in the folder to be read.
        It also creates an organized list with the names of the files to read.
        '''

        files = glob.glob(os.path.join(self.path, '*{}'.format(self.ext)))
        files = sorted(files)
        for f in files:
            filename = f.split('/')[-1]
            if folder_in_range(filename.split('.')[1], self.startDate, self.endDate, '%Y%m%d'):
                self.filenameList.append(f)

    def run(self, **kwargs):
        '''
        This method will be the one that will initiate the data entry, will be called constantly.
        You should first verify that your Setup () is set up and then continue to acquire
        the data to be processed with getData ().
        '''
        if not self.isConfig:
            self.setup(**kwargs)
            self.isConfig = True

        self.getData()

    def setup(self, 
              path=None,
              startDate=None,
              endDate=None,
              startTime=None,
              endTime=None,
              walk=True,
              code=None,
              online=False,
              mode=None,
              **kwargs):

        self.isConfig = True

        self.path = path
        self.startDate = startDate
        self.endDate = endDate
        self.startTime = startTime
        self.endTime = endTime
        self.walk = walk
        self.mode = int(mode)
        self.search_files()
        if self.filenameList:
            self.readFile()

    def getData(self):
        '''
        Before starting this function, you should check that there is still an unread file,
        If there are still blocks to read or if the data block is empty.

        You should call the file "read".

        '''

        if self.flagNoMoreFiles:
            self.dataOut.flagNoData = True
            raise schainpy.admin.SchainError('No more files')

        self.readBlock()

    def readFile(self):
        '''
        You must indicate if you are reading in Online or Offline mode and load the
        The parameters for this file reading mode.

        Then you must do 2 actions:

        1. Get the BLTR FileHeader.
        2. Start reading the first block.
        '''
        
        if self.fileSelector < len(self.filenameList):
            log.success('Opening file: {}'.format(self.filenameList[self.fileSelector]), self.name)
            self.fp = open(self.filenameList[self.fileSelector])
            self.fheader = FileHeaderBLTR(self.fp)
            self.rheader = RecordHeaderBLTR(self.fp)
            self.nFDTdataRecors = self.fheader.nFDTdataRecors
            self.fileSelector += 1    
            self.BlockCounter = 0
            return 1
        else:
            self.flagNoMoreFiles = True
            self.dataOut.flagNoData = True
            return 0

    def readBlock(self):
        '''
        It should be checked if the block has data, if it is not passed to the next file.

        Then the following is done:

        1. Read the RecordHeader
        2. Fill the buffer with the current block number.

        '''

        if self.BlockCounter == self.nFDTdataRecors:
            if not self.readFile():
                return
        
        if self.mode == 1:
            self.rheader.read(self.BlockCounter+1)
        elif self.mode == 0:
            self.rheader.read(self.BlockCounter)

        self.RecCounter = self.rheader.RecCounter
        self.OffsetStartHeader = self.rheader.OffsetStartHeader
        self.Off2StartNxtRec = self.rheader.Off2StartNxtRec
        self.Off2StartData = self.rheader.Off2StartData
        self.nProfiles = self.rheader.nProfiles
        self.nChannels = self.rheader.nChannels
        self.nHeights = self.rheader.nHeights
        self.frequency = self.rheader.TransmitFrec
        self.DualModeIndex = self.rheader.DualModeIndex
        self.pairsList = [(0, 1), (0, 2), (1, 2)]
        self.dataOut.pairsList = self.pairsList
        self.nRdPairs = len(self.dataOut.pairsList)
        self.dataOut.nRdPairs = self.nRdPairs
        self.dataOut.heightList = (self.rheader.StartRangeSamp + numpy.arange(self.nHeights) * self.rheader.SampResolution) / 1000.
        self.dataOut.channelList = range(self.nChannels)
        self.dataOut.nProfiles=self.rheader.nProfiles
        self.dataOut.nIncohInt=self.rheader.nIncohInt
        self.dataOut.nCohInt=self.rheader.nCohInt
        self.dataOut.ippSeconds= 1/float(self.rheader.PRFhz)
        self.dataOut.PRF=self.rheader.PRFhz
        self.dataOut.nFFTPoints=self.rheader.nProfiles
        self.dataOut.utctime = self.rheader.nUtime + self.rheader.nMilisec/1000.
        self.dataOut.timeZone = 0
        self.dataOut.useLocalTime = False
        self.dataOut.nmodes = 2
        log.log('Reading block {} - {}'.format(self.BlockCounter, self.dataOut.datatime), self.name)
        OffDATA = self.OffsetStartHeader + self.RecCounter * \
            self.Off2StartNxtRec + self.Off2StartData
        self.fp.seek(OffDATA, os.SEEK_SET)
            
        self.data_fft = numpy.fromfile(self.fp, [('complex','<c8')], self.nProfiles*self.nChannels*self.nHeights )
        self.data_fft = self.data_fft.astype(numpy.dtype('complex'))
        self.data_block = numpy.reshape(self.data_fft,(self.nHeights, self.nChannels, self.nProfiles))
        self.data_block = numpy.transpose(self.data_block, (1,2,0))
        copy = self.data_block.copy()
        spc = copy * numpy.conjugate(copy)
        self.data_spc = numpy.absolute(spc)  # valor absoluto o magnitud
        self.dataOut.data_spc = self.data_spc

        cspc = self.data_block.copy()
        self.data_cspc = self.data_block.copy()
        
        for i in range(self.nRdPairs):
            
            chan_index0 = self.dataOut.pairsList[i][0]
            chan_index1 = self.dataOut.pairsList[i][1]

            self.data_cspc[i, :, :] = cspc[chan_index0, :, :] * numpy.conjugate(cspc[chan_index1, :, :])

        '''Getting Eij and Nij'''
        (AntennaX0, AntennaY0) = pol2cart(
            self.rheader.AntennaCoord0, self.rheader.AntennaAngl0 * numpy.pi / 180)
        (AntennaX1, AntennaY1) = pol2cart(
            self.rheader.AntennaCoord1, self.rheader.AntennaAngl1 * numpy.pi / 180)
        (AntennaX2, AntennaY2) = pol2cart(
            self.rheader.AntennaCoord2, self.rheader.AntennaAngl2 * numpy.pi / 180)

        E01 = AntennaX0 - AntennaX1
        N01 = AntennaY0 - AntennaY1

        E02 = AntennaX0 - AntennaX2
        N02 = AntennaY0 - AntennaY2

        E12 = AntennaX1 - AntennaX2
        N12 = AntennaY1 - AntennaY2

        self.ChanDist = numpy.array(
            [[E01, N01], [E02, N02], [E12, N12]])

        self.dataOut.ChanDist = self.ChanDist

        self.BlockCounter += 2
        self.dataOut.data_spc = self.data_spc
        self.dataOut.data_cspc =self.data_cspc



'''
Created on Oct 24, 2016

@author: roj- LouVD
'''

import numpy
import copy
import datetime
import time
from time import gmtime

from numpy import transpose

from schainpy.model.proc.jroproc_base import ProcessingUnit, Operation, MPDecorator
from schainpy.model.data.jrodata import Parameters


class BLTRParametersProc(ProcessingUnit):    
    '''
    Processing unit for BLTR parameters data (winds)

    Inputs:
        self.dataOut.nmodes - Number of operation modes
        self.dataOut.nchannels - Number of channels
        self.dataOut.nranges - Number of ranges

        self.dataOut.data_SNR - SNR array 
        self.dataOut.data_output - Zonal, Vertical and Meridional velocity array
        self.dataOut.height - Height array (km)
        self.dataOut.time - Time array (seconds)

        self.dataOut.fileIndex -Index of the file currently read
        self.dataOut.lat - Latitude coordinate of BLTR location

        self.dataOut.doy - Experiment doy (number of the day in the current year) 
        self.dataOut.month - Experiment month
        self.dataOut.day - Experiment day
        self.dataOut.year - Experiment year
    '''

    def __init__(self):
        '''
        Inputs: None           
        '''
        ProcessingUnit.__init__(self)
        self.dataOut = Parameters()

    def setup(self, mode):
        '''
        '''
        self.dataOut.mode = mode

    def run(self, mode, snr_threshold=None):
        '''
        Inputs:
            mode = High resolution (0) or Low resolution (1) data
            snr_threshold = snr filter value
        '''

        if not self.isConfig:
            self.setup(mode)
            self.isConfig = True

        if self.dataIn.type == 'Parameters':
            self.dataOut.copy(self.dataIn)
        
        self.dataOut.data_param = self.dataOut.data[mode]
        self.dataOut.heightList = self.dataOut.height[0]
        self.dataOut.data_SNR = self.dataOut.data_SNR[mode]

        if snr_threshold is not None:
            SNRavg = numpy.average(self.dataOut.data_SNR, axis=0)
            SNRavgdB = 10*numpy.log10(SNRavg)
            for i in range(3):
                self.dataOut.data_param[i][SNRavgdB <= snr_threshold] = numpy.nan

# TODO

class OutliersFilter(Operation):

    def __init__(self):
        '''
        '''
        Operation.__init__(self)

    def run(self, svalue2, method, factor, filter, npoints=9):
        '''
        Inputs:
            svalue    -  string to select array velocity
            svalue2    -  string to choose axis filtering
            method    - 0 for SMOOTH or 1 for MEDIAN
            factor    - number used to set threshold
            filter    - 1 for data filtering using the standard deviation criteria else 0
            npoints    - number of points for mask filter
        '''

        print('    Outliers Filter {} {} / threshold = {}'.format(svalue, svalue, factor))

        
        yaxis = self.dataOut.heightList
        xaxis = numpy.array([[self.dataOut.utctime]])        

        # Zonal
        value_temp = self.dataOut.data_output[0]

        # Zonal
        value_temp = self.dataOut.data_output[1]
        
        # Vertical
        value_temp = numpy.transpose(self.dataOut.data_output[2])

        htemp = yaxis
        std = value_temp
        for h in range(len(htemp)):
            nvalues_valid = len(numpy.where(numpy.isfinite(value_temp[h]))[0])
            minvalid = npoints
            
            #only if valid values greater than the minimum required (10%)
            if nvalues_valid > minvalid:
                
                if method == 0:
                    #SMOOTH                    
                    w = value_temp[h] - self.Smooth(input=value_temp[h], width=npoints, edge_truncate=1)
                                            
                                          
                if method == 1:
                    #MEDIAN                        
                    w = value_temp[h] - self.Median(input=value_temp[h], width = npoints)                    
                                    
                dw = numpy.std(w[numpy.where(numpy.isfinite(w))],ddof = 1)                       
                
                threshold = dw*factor
                value_temp[numpy.where(w > threshold),h] = numpy.nan
                value_temp[numpy.where(w < -1*threshold),h] = numpy.nan

                                    
        #At the end
        if svalue2 == 'inHeight':
            value_temp = numpy.transpose(value_temp)
        output_array[:,m] = value_temp
        
        if svalue == 'zonal':
            self.dataOut.data_output[0] = output_array
           
        elif svalue == 'meridional':
            self.dataOut.data_output[1] = output_array

        elif svalue == 'vertical':
            self.dataOut.data_output[2] = output_array   
  
        return self.dataOut.data_output       

    
    def Median(self,input,width):
        '''
        Inputs:
            input - Velocity array
            width - Number of points for mask filter
            
        '''
        
        if numpy.mod(width,2) == 1:
            pc = int((width - 1) / 2)    
        cont = 0
        output = []
        
        for i in range(len(input)):        
            if i >= pc and i < len(input) - pc:
                new2 = input[i-pc:i+pc+1]            
                temp = numpy.where(numpy.isfinite(new2))
                new = new2[temp] 
                value = numpy.median(new)        
                output.append(value)
                
        output = numpy.array(output)
        output = numpy.hstack((input[0:pc],output))
        output = numpy.hstack((output,input[-pc:len(input)]))
        
        return output
    
    def Smooth(self,input,width,edge_truncate = None):
        '''
        Inputs:
            input - Velocity array
            width - Number of points for mask filter
            edge_truncate - 1 for truncate the convolution product else 
        
        '''        

        if numpy.mod(width,2) == 0:
            real_width = width + 1
            nzeros = width / 2
        else:
            real_width = width
            nzeros = (width - 1) / 2
    
        half_width = int(real_width)/2
        length = len(input)
        
        gate = numpy.ones(real_width,dtype='float')
        norm_of_gate = numpy.sum(gate)
        
        nan_process = 0
        nan_id = numpy.where(numpy.isnan(input))    
        if len(nan_id[0]) > 0:
            nan_process = 1
            pb = numpy.zeros(len(input))
            pb[nan_id] = 1.
            input[nan_id] = 0.
        
        if edge_truncate == True:
            output = numpy.convolve(input/norm_of_gate,gate,mode='same') 
        elif edge_truncate == False or edge_truncate == None:
            output = numpy.convolve(input/norm_of_gate,gate,mode='valid')
            output = numpy.hstack((input[0:half_width],output))
            output = numpy.hstack((output,input[len(input)-half_width:len(input)]))
                
        if nan_process:        
            pb = numpy.convolve(pb/norm_of_gate,gate,mode='valid') 
            pb = numpy.hstack((numpy.zeros(half_width),pb))
            pb = numpy.hstack((pb,numpy.zeros(half_width)))        
            output[numpy.where(pb > 0.9999)] = numpy.nan
            input[nan_id] = numpy.nan
        return output
    
    def Average(self,aver=0,nhaver=1): 
        '''
        Inputs:
            aver - Indicates the time period over which is averaged or consensus data
            nhaver - Indicates the decimation factor in heights  
        
        '''      
        nhpoints = 48 
                
        lat_piura = -5.17
        lat_huancayo = -12.04        
        lat_porcuya = -5.8
        
        if '%2.2f'%self.dataOut.lat == '%2.2f'%lat_piura:            
            hcm = 3.
            if self.dataOut.year == 2003 :            
                if self.dataOut.doy >= 25 and self.dataOut.doy < 64:
                    nhpoints = 12                
        
        elif '%2.2f'%self.dataOut.lat == '%2.2f'%lat_huancayo:            
            hcm = 3.
            if self.dataOut.year == 2003 :
                if self.dataOut.doy >= 25 and self.dataOut.doy < 64:
                    nhpoints = 12
                
            
        elif '%2.2f'%self.dataOut.lat == '%2.2f'%lat_porcuya:            
            hcm = 5.#2
        
        pdata = 0.2
        taver = [1,2,3,4,6,8,12,24]
        t0 = 0
        tf = 24        
        ntime =(tf-t0)/taver[aver]        
        ti = numpy.arange(ntime)
        tf = numpy.arange(ntime) + taver[aver]
        
        
        old_height = self.dataOut.heightList
        
        if nhaver > 1:
            num_hei = len(self.dataOut.heightList)/nhaver/self.dataOut.nmodes
            deltha = 0.05*nhaver
            minhvalid = pdata*nhaver
            for im in range(self.dataOut.nmodes):
                new_height = numpy.arange(num_hei)*deltha + self.dataOut.height[im,0] + deltha/2.
        
        
        data_fHeigths_List = []
        data_fZonal_List = []
        data_fMeridional_List = []
        data_fVertical_List = []
        startDTList = []
        

        for i in range(ntime): 
            height = old_height
            
            start = datetime.datetime(self.dataOut.year,self.dataOut.month,self.dataOut.day) + datetime.timedelta(hours = int(ti[i])) - datetime.timedelta(hours = 5)
            stop = datetime.datetime(self.dataOut.year,self.dataOut.month,self.dataOut.day) + datetime.timedelta(hours = int(tf[i])) - datetime.timedelta(hours = 5)
               
                        
            limit_sec1 = time.mktime(start.timetuple())
            limit_sec2 = time.mktime(stop.timetuple())
            
            t1 = numpy.where(self.f_timesec >= limit_sec1) 
            t2 = numpy.where(self.f_timesec < limit_sec2) 
            time_select = []
            for val_sec in t1[0]:
                if val_sec in t2[0]:
                    time_select.append(val_sec)
            
            
            time_select = numpy.array(time_select,dtype = 'int')
            minvalid = numpy.ceil(pdata*nhpoints)
            
            zon_aver = numpy.zeros([self.dataOut.nranges,self.dataOut.nmodes],dtype='f4') + numpy.nan
            mer_aver = numpy.zeros([self.dataOut.nranges,self.dataOut.nmodes],dtype='f4') + numpy.nan
            ver_aver = numpy.zeros([self.dataOut.nranges,self.dataOut.nmodes],dtype='f4') + numpy.nan
            
            if nhaver > 1:
                new_zon_aver = numpy.zeros([num_hei,self.dataOut.nmodes],dtype='f4') + numpy.nan
                new_mer_aver = numpy.zeros([num_hei,self.dataOut.nmodes],dtype='f4') + numpy.nan
                new_ver_aver = numpy.zeros([num_hei,self.dataOut.nmodes],dtype='f4') + numpy.nan
            
            if len(time_select) > minvalid:
                time_average = self.f_timesec[time_select]
                
                for im in range(self.dataOut.nmodes):
                    
                    for ih in range(self.dataOut.nranges):
                        if numpy.sum(numpy.isfinite(self.f_zon[time_select,ih,im])) >= minvalid:                            
                            zon_aver[ih,im] = numpy.nansum(self.f_zon[time_select,ih,im]) / numpy.sum(numpy.isfinite(self.f_zon[time_select,ih,im]))
                        
                        if numpy.sum(numpy.isfinite(self.f_mer[time_select,ih,im])) >= minvalid:
                            mer_aver[ih,im] = numpy.nansum(self.f_mer[time_select,ih,im]) / numpy.sum(numpy.isfinite(self.f_mer[time_select,ih,im]))
                        
                        if numpy.sum(numpy.isfinite(self.f_ver[time_select,ih,im])) >= minvalid:                            
                            ver_aver[ih,im] = numpy.nansum(self.f_ver[time_select,ih,im]) / numpy.sum(numpy.isfinite(self.f_ver[time_select,ih,im]))
                    
                    if nhaver > 1:
                        for ih in range(num_hei):
                            hvalid = numpy.arange(nhaver) + nhaver*ih
                            
                            if numpy.sum(numpy.isfinite(zon_aver[hvalid,im])) >= minvalid:
                                new_zon_aver[ih,im] = numpy.nansum(zon_aver[hvalid,im]) /  numpy.sum(numpy.isfinite(zon_aver[hvalid,im]))
                            
                            if numpy.sum(numpy.isfinite(mer_aver[hvalid,im])) >= minvalid:
                                new_mer_aver[ih,im] = numpy.nansum(mer_aver[hvalid,im]) /  numpy.sum(numpy.isfinite(mer_aver[hvalid,im]))
                            
                            if numpy.sum(numpy.isfinite(ver_aver[hvalid,im])) >= minvalid:
                                new_ver_aver[ih,im] = numpy.nansum(ver_aver[hvalid,im]) /  numpy.sum(numpy.isfinite(ver_aver[hvalid,im]))
                if nhaver > 1:
                    zon_aver = new_zon_aver
                    mer_aver = new_mer_aver
                    ver_aver = new_ver_aver
                    height = new_height
                
                                
                tstart = time_average[0]
                tend = time_average[-1]
                startTime = time.gmtime(tstart)
                                
                year = startTime.tm_year
                month = startTime.tm_mon
                day = startTime.tm_mday
                hour = startTime.tm_hour
                minute = startTime.tm_min            
                second = startTime.tm_sec
                
                startDTList.append(datetime.datetime(year,month,day,hour,minute,second))
                
                
                o_height = numpy.array([])
                o_zon_aver = numpy.array([])
                o_mer_aver = numpy.array([])
                o_ver_aver = numpy.array([])
                if self.dataOut.nmodes > 1:
                    for im in range(self.dataOut.nmodes):
                        
                        if im == 0:
                            h_select = numpy.where(numpy.bitwise_and(height[0,:] >=0,height[0,:] <= hcm,numpy.isfinite(height[0,:])))
                        else:                            
                            h_select = numpy.where(numpy.bitwise_and(height[1,:] > hcm,height[1,:] < 20,numpy.isfinite(height[1,:])))

                                                
                        ht = h_select[0]
                        
                        o_height = numpy.hstack((o_height,height[im,ht]))
                        o_zon_aver = numpy.hstack((o_zon_aver,zon_aver[ht,im]))
                        o_mer_aver = numpy.hstack((o_mer_aver,mer_aver[ht,im]))
                        o_ver_aver = numpy.hstack((o_ver_aver,ver_aver[ht,im]))

                    data_fHeigths_List.append(o_height)
                    data_fZonal_List.append(o_zon_aver)
                    data_fMeridional_List.append(o_mer_aver)
                    data_fVertical_List.append(o_ver_aver)
                        
 
                else:
                    h_select = numpy.where(numpy.bitwise_and(height[0,:] <= hcm,numpy.isfinite(height[0,:])))
                    ht = h_select[0]
                    o_height = numpy.hstack((o_height,height[im,ht]))
                    o_zon_aver = numpy.hstack((o_zon_aver,zon_aver[ht,im]))
                    o_mer_aver = numpy.hstack((o_mer_aver,mer_aver[ht,im]))
                    o_ver_aver = numpy.hstack((o_ver_aver,ver_aver[ht,im]))

                    data_fHeigths_List.append(o_height)
                    data_fZonal_List.append(o_zon_aver)
                    data_fMeridional_List.append(o_mer_aver)
                    data_fVertical_List.append(o_ver_aver)
                  
        
        return startDTList, data_fHeigths_List, data_fZonal_List, data_fMeridional_List, data_fVertical_List





'''
Created on Oct 24, 2016

@author: roj- LouVD
'''

import numpy
import datetime
import time
from time import gmtime

from numpy import transpose

from .jroproc_base import ProcessingUnit, Operation
from schainpy.model.data.jrodata import Parameters


class PXParametersProc(ProcessingUnit):    
    '''
    Processing unit for PX parameters data
    '''

    def __init__(self, **kwargs):
        """
        Inputs: None           
        """
        ProcessingUnit.__init__(self, **kwargs)
        self.dataOut = Parameters()
        self.isConfig = False

    def setup(self, mode):
        """
        """
        self.dataOut.mode = mode

    def run(self, mode):
        """
        Args:
            mode (str): select independent variable 'E' for elevation or 'A' for azimuth
        """

        if not self.isConfig:
            self.setup(mode)
            self.isConfig = True

        if self.dataIn.type == 'Parameters':
            self.dataOut.copy(self.dataIn)

        self.dataOut.data_param = numpy.array([self.dataOut.data[var] for var in self.dataOut.parameters])
        self.dataOut.data_param[self.dataOut.data_param == self.dataOut.missing] = numpy.nan

        if mode.upper()=='E':    
            self.dataOut.heightList = self.dataOut.data['Azimuth']
        else:
            self.dataOut.heightList = self.dataOut.data['Elevation']

        attrs = ['units', 'elevation', 'azimuth', 'max_range', 'latitude', 'longitude']
        meta = {}

        for attr in attrs:
            meta[attr] = getattr(self.dataOut, attr)

        meta['mode'] = mode
        self.dataOut.meta = meta


# Copyright (c) 2012-2020 Jicamarca Radio Observatory
# All rights reserved.
#
# Distributed under the terms of the BSD 3-clause license.
"""Base class to create plot operations

"""

import os
import sys
import zmq
import time
import numpy
import datetime
from multiprocessing import Queue
from functools import wraps
from threading import Thread
import matplotlib

if 'BACKEND' in os.environ:
    matplotlib.use(os.environ['BACKEND'])
elif 'linux' in sys.platform:
    matplotlib.use("TkAgg")
elif 'darwin' in sys.platform:
    matplotlib.use('WxAgg')
else:
    from schainpy.utils import log
    log.warning('Using default Backend="Agg"', 'INFO')
    matplotlib.use('Agg')

import matplotlib.pyplot as plt
from matplotlib.patches import Polygon
from mpl_toolkits.axes_grid1 import make_axes_locatable
from matplotlib.ticker import FuncFormatter, LinearLocator, MultipleLocator

from schainpy.model.data.jrodata import PlotterData
from schainpy.model.proc.jroproc_base import ProcessingUnit, Operation, MPDecorator
from schainpy.utils import log

jet_values = matplotlib.pyplot.get_cmap('jet', 100)(numpy.arange(100))[10:90]
blu_values = matplotlib.pyplot.get_cmap(
    'seismic_r', 20)(numpy.arange(20))[10:15]
ncmap = matplotlib.colors.LinearSegmentedColormap.from_list(
    'jro', numpy.vstack((blu_values, jet_values)))
matplotlib.pyplot.register_cmap(cmap=ncmap)

CMAPS = [plt.get_cmap(s) for s in ('jro', 'jet', 'viridis',
                                   'plasma', 'inferno', 'Greys', 'seismic', 'bwr', 'coolwarm')]

EARTH_RADIUS = 6.3710e3

def ll2xy(lat1, lon1, lat2, lon2):

    p = 0.017453292519943295
    a = 0.5 - numpy.cos((lat2 - lat1) * p)/2 + numpy.cos(lat1 * p) * \
        numpy.cos(lat2 * p) * (1 - numpy.cos((lon2 - lon1) * p)) / 2
    r = 12742 * numpy.arcsin(numpy.sqrt(a))
    theta = numpy.arctan2(numpy.sin((lon2-lon1)*p)*numpy.cos(lat2*p), numpy.cos(lat1*p)
                          * numpy.sin(lat2*p)-numpy.sin(lat1*p)*numpy.cos(lat2*p)*numpy.cos((lon2-lon1)*p))
    theta = -theta + numpy.pi/2
    return r*numpy.cos(theta), r*numpy.sin(theta)


def km2deg(km):
    '''
    Convert distance in km to degrees
    '''

    return numpy.rad2deg(km/EARTH_RADIUS)


def figpause(interval):
    backend = plt.rcParams['backend']
    if backend in matplotlib.rcsetup.interactive_bk:
        figManager = matplotlib._pylab_helpers.Gcf.get_active()
        if figManager is not None:
            canvas = figManager.canvas
            if canvas.figure.stale:
                canvas.draw()
            try:
                canvas.start_event_loop(interval)
            except:
                pass
            return


def popup(message):
    '''
    '''

    fig = plt.figure(figsize=(12, 8), facecolor='r')
    text = '\n'.join([s.strip() for s in message.split(':')])
    fig.text(0.01, 0.5, text, ha='left', va='center',
             size='20', weight='heavy', color='w')
    fig.show()
    figpause(1000)


class Throttle(object):
    '''
    Decorator that prevents a function from being called more than once every
    time period.
    To create a function that cannot be called more than once a minute, but
    will sleep until it can be called:
    @Throttle(minutes=1)
    def foo():
      pass

    for i in range(10):
      foo()
      print "This function has run %s times." % i
    '''

    def __init__(self, seconds=0, minutes=0, hours=0):
        self.throttle_period = datetime.timedelta(
            seconds=seconds, minutes=minutes, hours=hours
        )

        self.time_of_last_call = datetime.datetime.min

    def __call__(self, fn):
        @wraps(fn)
        def wrapper(*args, **kwargs):
            coerce = kwargs.pop('coerce', None)
            if coerce:
                self.time_of_last_call = datetime.datetime.now()
                return fn(*args, **kwargs)
            else:
                now = datetime.datetime.now()
                time_since_last_call = now - self.time_of_last_call
                time_left = self.throttle_period - time_since_last_call

                if time_left > datetime.timedelta(seconds=0):
                    return

            self.time_of_last_call = datetime.datetime.now()
            return fn(*args, **kwargs)

        return wrapper

def apply_throttle(value):

    @Throttle(seconds=value)
    def fnThrottled(fn):
        fn()

    return fnThrottled


@MPDecorator
class Plot(Operation):
    """Base class for Schain plotting operations

    This class should never be use directtly you must subclass a new operation,
    children classes must be defined as follow:

    ExamplePlot(Plot):

        CODE = 'code'
        colormap = 'jet'
        plot_type = 'pcolor' # options are ('pcolor', 'pcolorbuffer', 'scatter', 'scatterbuffer')

        def setup(self):
            pass

        def plot(self):
            pass

    """

    CODE = 'Figure'
    colormap = 'jet'
    bgcolor = 'white'
    buffering = True
    __missing = 1E30

    __attrs__ = ['show', 'save', 'ymin', 'ymax', 'zmin', 'zmax', 'title',
                 'showprofile']

    def __init__(self):

        Operation.__init__(self)
        self.isConfig = False
        self.isPlotConfig = False
        self.save_time = 0
        self.sender_time = 0
        self.data = None
        self.firsttime = True
        self.sender_queue = Queue(maxsize=60)
        self.plots_adjust = {'left': 0.125, 'right': 0.9, 'bottom': 0.15, 'top': 0.9, 'wspace': 0.2, 'hspace': 0.2}

    def __fmtTime(self, x, pos):
        '''
        '''

        return '{}'.format(self.getDateTime(x).strftime('%H:%M'))

    def __setup(self, **kwargs):
        '''
        Initialize variables
        '''

        self.figures = []
        self.axes = []
        self.cb_axes = []
        self.localtime = kwargs.pop('localtime', True)
        self.show = kwargs.get('show', True)
        self.save = kwargs.get('save', False)
        self.save_period = kwargs.get('save_period', 60)
        self.colormap = kwargs.get('colormap', self.colormap)
        self.colormap_coh = kwargs.get('colormap_coh', 'jet')
        self.colormap_phase = kwargs.get('colormap_phase', 'RdBu_r')
        self.colormaps = kwargs.get('colormaps', None)
        self.bgcolor = kwargs.get('bgcolor', self.bgcolor)
        self.showprofile = kwargs.get('showprofile', False)
        self.title = kwargs.get('wintitle', self.CODE.upper())
        self.cb_label = kwargs.get('cb_label', None)
        self.cb_labels = kwargs.get('cb_labels', None)
        self.labels = kwargs.get('labels', None)
        self.xaxis = kwargs.get('xaxis', 'frequency')
        self.zmin = kwargs.get('zmin', None)
        self.zmax = kwargs.get('zmax', None)
        self.zlimits = kwargs.get('zlimits', None)
        self.xmin = kwargs.get('xmin', None)
        self.xmax = kwargs.get('xmax', None)
        self.xrange = kwargs.get('xrange', 12)
        self.xscale = kwargs.get('xscale', None)
        self.ymin = kwargs.get('ymin', None)
        self.ymax = kwargs.get('ymax', None)
        self.yscale = kwargs.get('yscale', None)
        self.xlabel = kwargs.get('xlabel', None)
        self.attr_time = kwargs.get('attr_time', 'utctime')
        self.decimation = kwargs.get('decimation', None)
        self.showSNR = kwargs.get('showSNR', False)
        self.oneFigure = kwargs.get('oneFigure', True)
        self.width = kwargs.get('width', None)
        self.height = kwargs.get('height', None)
        self.colorbar = kwargs.get('colorbar', True)
        self.factors = kwargs.get('factors', [1, 1, 1, 1, 1, 1, 1, 1])
        self.channels = kwargs.get('channels', None)
        self.titles = kwargs.get('titles', [])
        self.polar = False
        self.type = kwargs.get('type', 'iq')
        self.grid = kwargs.get('grid', False)
        self.pause = kwargs.get('pause', False)
        self.save_code = kwargs.get('save_code', self.CODE)
        self.throttle = kwargs.get('throttle', 0)
        self.exp_code = kwargs.get('exp_code', None)
        self.plot_server = kwargs.get('plot_server', False)
        self.sender_period = kwargs.get('sender_period', 60)
        self.tag = kwargs.get('tag', '')
        self.height_index = kwargs.get('height_index', None)
        self.__throttle_plot = apply_throttle(self.throttle)
        self.data = PlotterData(
            self.CODE, self.throttle, self.exp_code, self.localtime, self.buffering, snr=self.showSNR)
        
        if self.plot_server:
            if not self.plot_server.startswith('tcp://'):
                self.plot_server = 'tcp://{}'.format(self.plot_server)
            log.success(
                'Sending to server: {}'.format(self.plot_server),
                self.name
            )

    def __setup_plot(self):
        '''
        Common setup for all figures, here figures and axes are created
        '''

        self.setup()

        self.time_label = 'LT' if self.localtime else 'UTC'        

        if self.width is None:
            self.width = 8

        self.figures = []
        self.axes = []
        self.cb_axes = []
        self.pf_axes = []
        self.cmaps = []

        size = '15%' if self.ncols == 1 else '30%'
        pad = '4%' if self.ncols == 1 else '8%'

        if self.oneFigure:
            if self.height is None:
                self.height = 1.4 * self.nrows + 1
            fig = plt.figure(figsize=(self.width, self.height),
                             edgecolor='k',
                             facecolor='w')
            self.figures.append(fig)
            for n in range(self.nplots):
                ax = fig.add_subplot(self.nrows, self.ncols,
                                     n + 1, polar=self.polar)
                ax.tick_params(labelsize=8)
                ax.firsttime = True
                ax.index = 0
                ax.press = None
                self.axes.append(ax)
                if self.showprofile:
                    cax = self.__add_axes(ax, size=size, pad=pad)
                    cax.tick_params(labelsize=8)
                    self.pf_axes.append(cax)
        else:
            if self.height is None:
                self.height = 3
            for n in range(self.nplots):
                fig = plt.figure(figsize=(self.width, self.height),
                                 edgecolor='k',
                                 facecolor='w')
                ax = fig.add_subplot(1, 1, 1, polar=self.polar)
                ax.tick_params(labelsize=8)
                ax.firsttime = True
                ax.index = 0
                ax.press = None
                self.figures.append(fig)
                self.axes.append(ax)
                if self.showprofile:
                    cax = self.__add_axes(ax, size=size, pad=pad)
                    cax.tick_params(labelsize=8)
                    self.pf_axes.append(cax)

        for n in range(self.nrows):
            if self.colormaps is not None:
                cmap = plt.get_cmap(self.colormaps[n])
            else:
                cmap = plt.get_cmap(self.colormap)
            cmap.set_bad(self.bgcolor, 1.)
            self.cmaps.append(cmap)

    def __add_axes(self, ax, size='30%', pad='8%'):
        '''
        Add new axes to the given figure
        '''
        divider = make_axes_locatable(ax)
        nax = divider.new_horizontal(size=size, pad=pad)
        ax.figure.add_axes(nax)
        return nax

    def fill_gaps(self, x_buffer, y_buffer, z_buffer):
        '''
        Create a masked array for missing data
        '''
        if x_buffer.shape[0] < 2:
            return x_buffer, y_buffer, z_buffer

        deltas = x_buffer[1:] - x_buffer[0:-1]
        x_median = numpy.median(deltas)

        index = numpy.where(deltas > 5 * x_median)

        if len(index[0]) != 0:
            z_buffer[::, index[0], ::] = self.__missing
            z_buffer = numpy.ma.masked_inside(z_buffer,
                                              0.99 * self.__missing,
                                              1.01 * self.__missing)

        return x_buffer, y_buffer, z_buffer

    def decimate(self):

        # dx = int(len(self.x)/self.__MAXNUMX) + 1
        dy = int(len(self.y) / self.decimation) + 1

        # x = self.x[::dx]
        x = self.x
        y = self.y[::dy]
        z = self.z[::, ::, ::dy]

        return x, y, z

    def format(self):
        '''
        Set min and max values, labels, ticks and titles
        '''
            
        for n, ax in enumerate(self.axes):
            if ax.firsttime:
                if self.xaxis != 'time':
                    xmin = self.xmin
                    xmax = self.xmax
                else:
                    xmin = self.tmin
                    xmax = self.tmin + self.xrange*60*60
                    ax.xaxis.set_major_formatter(FuncFormatter(self.__fmtTime))
                    ax.xaxis.set_major_locator(LinearLocator(9))
                ymin = self.ymin if self.ymin else numpy.nanmin(self.y)
                ymax = self.ymax if self.ymax else numpy.nanmax(self.y)
                ax.set_facecolor(self.bgcolor)
                if self.xscale:
                    ax.xaxis.set_major_formatter(FuncFormatter(
                        lambda x, pos: '{0:g}'.format(x*self.xscale)))
                if self.yscale:
                    ax.yaxis.set_major_formatter(FuncFormatter(
                        lambda x, pos: '{0:g}'.format(x*self.yscale)))
                if self.xlabel is not None:
                    ax.set_xlabel(self.xlabel)
                if self.ylabel is not None:
                    ax.set_ylabel(self.ylabel)
                if self.showprofile:
                    self.pf_axes[n].set_ylim(ymin, ymax)
                    self.pf_axes[n].set_xlim(self.zmin, self.zmax)
                    self.pf_axes[n].set_xlabel('dB')
                    self.pf_axes[n].grid(b=True, axis='x')
                    [tick.set_visible(False)
                     for tick in self.pf_axes[n].get_yticklabels()]
                if self.colorbar:
                    ax.cbar = plt.colorbar(
                        ax.plt, ax=ax, fraction=0.05, pad=0.02, aspect=10)
                    ax.cbar.ax.tick_params(labelsize=8)
                    ax.cbar.ax.press = None
                    if self.cb_label:
                        ax.cbar.set_label(self.cb_label, size=8)
                    elif self.cb_labels:
                        ax.cbar.set_label(self.cb_labels[n], size=8)
                else:
                    ax.cbar = None
                ax.set_xlim(xmin, xmax)
                ax.set_ylim(ymin, ymax)
                ax.firsttime = False
                if self.grid:
                    ax.grid(True)
            if not self.polar:
                ax.set_title('{} {} {}'.format(
                    self.titles[n],
                    self.getDateTime(self.data.max_time).strftime(
                        '%Y-%m-%d %H:%M:%S'),
                    self.time_label),
                    size=8)
            else:
                ax.set_title('{}'.format(self.titles[n]), size=8)
                ax.set_ylim(0, 90)
                ax.set_yticks(numpy.arange(0, 90, 20))
                ax.yaxis.labelpad = 40

        if self.firsttime:
            for n, fig in enumerate(self.figures):
                fig.subplots_adjust(**self.plots_adjust)
            self.firsttime = False

    def clear_figures(self):
        '''
        Reset axes for redraw plots
        '''

        for ax in self.axes+self.pf_axes+self.cb_axes:
            ax.clear()
            ax.firsttime = True
            if hasattr(ax, 'cbar') and ax.cbar:
                ax.cbar.remove()

    def __plot(self):
        '''
        Main function to plot, format and save figures
        '''

        self.plot()
        self.format()
        
        for n, fig in enumerate(self.figures):
            if self.nrows == 0 or self.nplots == 0:
                log.warning('No data', self.name)
                fig.text(0.5, 0.5, 'No Data', fontsize='large', ha='center')
                fig.canvas.manager.set_window_title(self.CODE)
                continue
            
            fig.canvas.manager.set_window_title('{} - {}'.format(self.title,
                                                                 self.getDateTime(self.data.max_time).strftime('%Y/%m/%d')))
            fig.canvas.draw()
            if self.show:
                fig.show()
                figpause(0.01)

            if self.save:
                self.save_figure(n)
        
        if self.plot_server:
            self.send_to_server()

    def save_figure(self, n):
        '''
        '''

        if (self.data.tm - self.save_time) < self.sender_period:
            return

        self.save_time = self.data.tm

        fig = self.figures[n]

        figname = os.path.join(
            self.save,
            self.save_code,
            '{}_{}.png'.format(                
                self.save_code,
                self.getDateTime(self.data.max_time).strftime(
                    '%Y%m%d_%H%M%S'
                    ),
                )
            )
        log.log('Saving figure: {}'.format(figname), self.name)
        if not os.path.isdir(os.path.dirname(figname)):
            os.makedirs(os.path.dirname(figname))
        fig.savefig(figname)

        if self.throttle == 0:
            figname = os.path.join(
                self.save,
                '{}_{}.png'.format(
                    self.save_code,
                    self.getDateTime(self.data.min_time).strftime(
                        '%Y%m%d'
                        ),
                    )
                )
            fig.savefig(figname)

    def send_to_server(self):
        '''
        '''

        interval = self.data.tm - self.sender_time
        if interval < self.sender_period:
            return

        self.sender_time = self.data.tm
        
        attrs = ['titles', 'zmin', 'zmax', 'tag', 'ymin', 'ymax']
        for attr in attrs:
            value = getattr(self, attr)
            if value:
                if isinstance(value, (numpy.float32, numpy.float64)):
                    value = round(float(value), 2)
                self.data.meta[attr] = value
        if self.colormap == 'jet':
            self.data.meta['colormap'] = 'Jet'
        elif 'RdBu' in self.colormap:
            self.data.meta['colormap'] = 'RdBu'
        else:
            self.data.meta['colormap'] = 'Viridis'
        self.data.meta['interval'] = int(interval)

        try:
            self.sender_queue.put(self.data.tm, block=False)
        except:
            tm = self.sender_queue.get()
            self.sender_queue.put(self.data.tm)
        
        while True:
            if self.sender_queue.empty():
                break
            tm = self.sender_queue.get()
            try:
                msg = self.data.jsonify(tm, self.save_code, self.plot_type)
            except:
                continue
            self.socket.send_string(msg)
            socks = dict(self.poll.poll(5000))
            if socks.get(self.socket) == zmq.POLLIN:
                reply = self.socket.recv_string()
                if reply == 'ok':
                    log.log("Response from server ok", self.name)
                    time.sleep(0.2)
                    continue
                else:
                    log.warning(
                        "Malformed reply from server: {}".format(reply), self.name)
            else:
                log.warning(
                    "No response from server, retrying...", self.name)
                self.sender_queue.put(self.data.tm)
            self.socket.setsockopt(zmq.LINGER, 0)
            self.socket.close()
            self.poll.unregister(self.socket)
            time.sleep(0.1)
            self.socket = self.context.socket(zmq.REQ)
            self.socket.connect(self.plot_server)
            self.poll.register(self.socket, zmq.POLLIN)
            break

    def setup(self):
        '''
        This method should be implemented in the child class, the following
        attributes should be set:

        self.nrows: number of rows
        self.ncols: number of cols
        self.nplots: number of plots (channels or pairs)
        self.ylabel: label for Y axes
        self.titles: list of axes title 

        '''
        raise NotImplementedError

    def plot(self):
        '''
        Must be defined in the child class
        '''
        raise NotImplementedError
    
    def run(self, dataOut, **kwargs):
        '''
        Main plotting routine
        '''

        if self.isConfig is False:
            self.__setup(**kwargs)

            if self.localtime:
                self.getDateTime = datetime.datetime.fromtimestamp
            else:
                self.getDateTime = datetime.datetime.utcfromtimestamp

            self.data.setup()
            self.isConfig = True
            if self.plot_server:
                self.context = zmq.Context()
                self.socket = self.context.socket(zmq.REQ)
                self.socket.connect(self.plot_server)
                self.poll = zmq.Poller()
                self.poll.register(self.socket, zmq.POLLIN)

        tm = getattr(dataOut, self.attr_time)
        
        if self.data and 'time' in self.xaxis and (tm - self.tmin) >= self.xrange*60*60:    
            self.save_counter = self.save_period
            self.__plot()
            self.tmin += self.xrange*60*60
            self.data.setup()
            self.clear_figures()

        self.data.update(dataOut, tm)

        if self.isPlotConfig is False:
            self.__setup_plot()
            self.isPlotConfig = True
            if self.xaxis == 'time':
                dt = self.getDateTime(tm)
                if self.xmin is None:
                    self.tmin = tm
                    self.xmin = dt.hour    
                minutes = (self.xmin-int(self.xmin)) * 60
                seconds = (minutes - int(minutes)) * 60
                self.tmin = (dt.replace(hour=int(self.xmin), minute=int(minutes), second=int(seconds)) -
                        datetime.datetime(1970, 1, 1)).total_seconds()
                if self.localtime:
                    self.tmin += time.timezone

                if self.xmin is not None and self.xmax is not None:
                    self.xrange = self.xmax - self.xmin

        if self.throttle == 0:
            self.__plot()
        else:
            self.__throttle_plot(self.__plot)#, coerce=coerce)

    def close(self):

        if self.data and not self.data.flagNoData:
            self.save_counter = self.save_period
            self.__plot()
        if self.data and not self.data.flagNoData and self.pause:
            figpause(10)




"""
The admin module contains all administrative classes relating to the schain python api.

The main role of this module is to send some reports. It contains a
notification class and a standard error handing class.

$Id: admin.py 3966 2015-12-01 14:32:29Z miguel.urco $
"""
import os
import sys
import time
import traceback
import smtplib
if sys.version[0] == '3':
    from configparser import ConfigParser
else:
    from ConfigParser import ConfigParser
import io
from threading import Thread
from multiprocessing import Process
from email.mime.text import MIMEText
from email.mime.application import MIMEApplication
from email.mime.multipart import MIMEMultipart

import schainpy
from schainpy.utils import log
from schainpy.model.graphics.jroplot_base import popup

def get_path():
    '''
    Return schainpy path
    '''
    
    try:
        root = __file__
        if os.path.islink(root):
            root = os.path.realpath(root)
            
        return os.path.dirname(os.path.abspath(root))
    except:
        log.error('I am sorry, but something is wrong... __file__ not found')

class Alarm(Process):
    '''
    modes:
      0 - All
      1 - Send email
      2 - Popup message
      3 - Sound alarm
      4 - Send to alarm system TODO
    '''

    def __init__(self, modes=[], **kwargs):
        Process.__init__(self)
        self.modes = modes
        self.kwargs = kwargs

    @staticmethod
    def play_sound():
        sound = os.path.join(get_path(), 'alarm1.oga')
        if os.path.exists(sound):        
            for __ in range(2):
                os.system('paplay {}'.format(sound))
                time.sleep(0.5)
        else:
            log.warning('Unable to play alarm, sound file not found', 'ADMIN')
    
    @staticmethod
    def send_email(**kwargs):
        notifier = SchainNotify()
        notifier.notify(**kwargs)            

    @staticmethod
    def show_popup(message):
        if isinstance(message, list):
            message = message[-1]
        popup(message)

    @staticmethod
    def send_alarm():
        pass

    @staticmethod
    def get_kwargs(kwargs, keys):
        ret = {}
        for key in keys:
            ret[key] = kwargs[key]
        return ret

    def run(self):
        tasks = {
            1 : self.send_email,
            2 : self.show_popup,
            3 : self.play_sound,
            4 : self.send_alarm,
        }

        tasks_args = {
            1: ['email', 'message', 'subject', 'subtitle', 'filename'],
            2: ['message'],
            3: [],
            4: [],
        }
        procs = []
        for mode in self.modes:
            if 0 in self.modes:
                for x in tasks:
                    t = Thread(target=tasks[x], kwargs=self.get_kwargs(self.kwargs, tasks_args[x]))
                    t.start()
                    procs.append(t)
                break
            else:
                t = Thread(target=tasks[mode], kwargs=self.get_kwargs(self.kwargs, tasks_args[mode]))
                t.start()
                procs.append(t)
        for t in procs:
            t.join()


class SchainConfigure():
    
    __DEFAULT_ADMINISTRATOR_EMAIL = "juan.espinoza@jro.igp.gob.pe"
    __DEFAULT_EMAIL_SERVER = "jro-zimbra.igp.gob.pe"
    __DEFAULT_SENDER_EMAIL = "notifier-schain@jro.igp.gob.pe"
    __DEFAULT_SENDER_PASS = ""
    
    __SCHAIN_ADMINISTRATOR_EMAIL = "CONTACT"
    __SCHAIN_EMAIL_SERVER = "MAILSERVER"
    __SCHAIN_SENDER_EMAIL = "MAILSERVER_ACCOUNT"
    __SCHAIN_SENDER_PASS = "MAILSERVER_PASSWORD"
    
    def __init__(self, initFile = None):
        
        # Set configuration file
        if (initFile == None):
            self.__confFilePath = "/etc/schain.conf"
        else:
            self.__confFilePath = initFile

        # open configuration file
        try:
            self.__confFile = open(self.__confFilePath, "r")
        except IOError:
            # can't read from file - use all hard-coded values
            self.__initFromHardCode()
            return
        
        # create Parser using standard module ConfigParser
        self.__parser = ConfigParser()
        
        # read conf file into a StringIO with "[madrigal]\n" section heading prepended
        strConfFile = io.StringIO("[schain]\n" + self.__confFile.read())

        # parse StringIO configuration file
        self.__parser.readfp(strConfFile)
        
        # read information from configuration file
        self.__readConfFile()

        # close conf file
        self.__confFile.close()
    
    
    def __initFromHardCode(self):
        
        self.__sender_email = self.__DEFAULT_SENDER_EMAIL
        self.__sender_pass = self.__DEFAULT_SENDER_PASS
        self.__admin_email = self.__DEFAULT_ADMINISTRATOR_EMAIL
        self.__email_server = self.__DEFAULT_EMAIL_SERVER
    
    def __readConfFile(self):
        """__readConfFile is a private helper function that reads information from the parsed config file.
        
        Inputs: None
        
        Returns: Void.

        Affects: Initializes class member variables that are found in the config file.

        Exceptions: MadrigalError thrown if any key not found.
        """

        # get the sender email
        try:
            self.__sender_email = self.__parser.get("schain", self.__SCHAIN_SENDER_EMAIL)
        except:
            self.__sender_email = self.__DEFAULT_SENDER_EMAIL
        
        # get the sender password
        try:
            self.__sender_pass = self.__parser.get("schain", self.__SCHAIN_SENDER_PASS)
        except:
            self.__sender_pass = self.__DEFAULT_SENDER_PASS
            
        # get the administrator email
        try:
            self.__admin_email = self.__parser.get("schain", self.__SCHAIN_ADMINISTRATOR_EMAIL)
        except:
            self.__admin_email = self.__DEFAULT_ADMINISTRATOR_EMAIL
            
        # get the server email
        try:
            self.__email_server = self.__parser.get("schain", self.__SCHAIN_EMAIL_SERVER)
        except:
            self.__email_server = self.__DEFAULT_EMAIL_SERVER
    
    def getEmailServer(self):
        
        return self.__email_server
    
    def getSenderEmail(self):
        
        return self.__sender_email
    
    def getSenderPass(self):
        
        return self.__sender_pass
    
    def getAdminEmail(self):
        
        return self.__admin_email
    
class SchainNotify:
    """SchainNotify is an object used to send messages to an administrator about a Schain software.

    This object provides functions needed to send messages to an administrator about a Schain , for now
    only sendAlert, which sends an email to the site administrator found is ADMIN_EMAIL

    Usage example:

        import schainpy.admin
    
        try:
        
            adminObj =  schainpy.admin.SchainNotify()
            adminObj.sendAlert('This is important!', 'Important Message')
            
        except schainpy.admin.SchainError, e:
        
            print e.getExceptionStr()


    Non-standard Python modules used:
    None

    Exceptions thrown: None - Note that SchainNotify tries every trick it knows to avoid
    throwing exceptions, since this is the class that will generally be called when there is a problem.

    Change history:

    Written by "Miguel Urco":mailto:miguel.urco@jro.igp.gob.pe  Dec. 1, 2015
    """

    #constants
    
    def __init__(self):
        """__init__ initializes SchainNotify by getting some basic information from SchainDB and SchainSite.

        Note that SchainNotify tries every trick it knows to avoid throwing exceptions, since
        this is the class that will generally be called when there is a problem.

        Inputs: Existing SchainDB object, by default = None.
        
        Returns: void

        Affects: Initializes self.__binDir.

        Exceptions: None.
        """
    
        # note that the main configuration file is unavailable 
        # the best that can be done is send an email to root using localhost mailserver
        confObj = SchainConfigure()
        
        self.__emailFromAddress = confObj.getSenderEmail()
        self.__emailPass = confObj.getSenderPass()
        self.__emailToAddress = confObj.getAdminEmail()
        self.__emailServer  = confObj.getEmailServer()

    def sendEmail(self, email_from, email_to, subject='Error running ...', message="", subtitle="", filename="", html_format=True):

        if not email_to:
            return 0
        
        if not self.__emailServer:
            return 0
        
        log.success('Sending email to {}...'.format(email_to), 'System')

        msg = MIMEMultipart()
        msg['Subject'] = subject
        msg['From'] = "SChain API (v{}) <{}>".format(schainpy.__version__, email_from)
        msg['Reply-to'] = email_from
        msg['To'] = email_to
        
        # That is what u see if dont have an email reader:
        msg.preamble = 'SChainPy'
        
        if html_format:
            message = "<h1> %s </h1>" %subject + "<h3>" + subtitle.replace("\n", "</h3><h3>\n") + "</h3>" + message.replace("\n", "<br>\n")
            message = "<html>\n" + message + '</html>'
        
            # This is the textual part:
            part = MIMEText(message, "html")
        else:
            message = subject + "\n" + subtitle + "\n" + message
            part = MIMEText(message)
            
        msg.attach(part)
        
        if filename and os.path.isfile(filename):
            # This is the binary part(The Attachment):
            part = MIMEApplication(open(filename,"rb").read())
            part.add_header('Content-Disposition', 
                            'attachment',
                            filename=os.path.basename(filename))
            msg.attach(part)
        
        # Create an instance in SMTP server
        try:
            smtp = smtplib.SMTP(self.__emailServer)
        except:
            log.error('Could not connect to server {}'.format(self.__emailServer), 'System')
            return 0
            
        # Start the server:
    #         smtp.ehlo()
        if self.__emailPass:
            smtp.login(self.__emailFromAddress, self.__emailPass)
        
        # Send the email
        try:            
            smtp.sendmail(msg['From'], msg['To'], msg.as_string())
        except:
            log.error('Could not send the email to {}'.format(msg['To']), 'System')
            smtp.quit()
            return 0
        
        smtp.quit()

        log.success('Email sent ', 'System')
        
        return 1
    
    def sendAlert(self, message, subject = "", subtitle="", filename=""):
        """sendAlert sends an email with the given message and optional title.

        Inputs: message (string), and optional title (string)
        
        Returns: void

        Affects: none

        Exceptions: None.
        """
        
        if not self.__emailToAddress:
            return 0
        
        print("***** Sending alert to %s *****" %self.__emailToAddress)
        # set up message
    
        sent=self.sendEmail(email_from=self.__emailFromAddress,
                           email_to=self.__emailToAddress,
                           subject=subject,
                           message=message,
                           subtitle=subtitle, 
                           filename=filename)
        
        if not sent:
            return 0
        
        return 1
        
    def notify(self, email, message, subject = "", subtitle="", filename=""):
        """notify sends an email with the given message and title to email.

        Inputs: email (string), message (string), and subject (string)
        
        Returns: void

        Affects: none

        Exceptions: None.
        """
        
        if email is None:
            email = self.__emailToAddress
        
        self.sendEmail(
            email_from=self.__emailFromAddress,
            email_to=email,
            subject=subject,
            message=message,
            subtitle=subtitle, 
            filename=filename
            )


class SchainError(Exception):
    """SchainError is an exception class that is thrown for all known errors using Schain Py lib.

    Usage example:

        import sys, traceback
        import schainpy.admin
    
        try:
        
            test = open('ImportantFile.txt', 'r')
            
        except:
        
            raise schainpy.admin.SchainError('ImportantFile.txt not opened!',
                                                traceback.format_exception(sys.exc_info()[0],
                                                                        sys.exc_info()[1],
                                                                        sys.exc_info()[2]))
    """


    def __init__(self, strInterpretation, exceptionList=None):
        """ __init__ gathers the interpretation string along with all information from sys.exc_info().

        Inputs:
                strIntepretation - A string representing the programmer's interpretation of
                why the exception occurred

                exceptionList - a list of strings completely describing the exception.
                Generated by traceback.format_exception(sys.exc_info()[0],
                                                        sys.exc_info()[1],
                                                        sys.exc_info()[2])
        
        Returns: Void.

        Affects: Initializes class member variables _strInterp, _strExcList.

        Exceptions: None.
        """
        
        if not exceptionList:
            exceptionList = traceback.format_exception(sys.exc_info()[0],
                                                        sys.exc_info()[1],
                                                        sys.exc_info()[2])
            
        self._strInterp = strInterpretation
        self._strExcList = exceptionList

        
    def getExceptionStr(self):
        """ getExceptionStr returns a formatted string ready for printing completely describing the exception.

        Inputs: None
        
        Returns: A formatted string ready for printing completely describing the exception.

        Affects: None

        Exceptions: None.
        """
        excStr = ''
        excStr = excStr + self._strInterp + '\n\n'

        if self._strExcList != None:
            for item in self._strExcList:
                excStr = excStr + str(item) + '\n'

        return excStr
    
    def __str__(self):
        
        return(self.getExceptionStr())


    def getExceptionHtml(self):
        """ getExceptionHtml returns an Html formatted string completely describing the exception.

        Inputs: None
        
        Returns: A formatted string ready for printing completely describing the exception.

        Affects: None

        Exceptions: None.
        """
        
        excStr = '<BR>The following Schain Python exception has occurred:\n<BR>'
        excStr = excStr + self._strInterp + '\n<BR>\n'

        if self._strExcList != None:
            for item in self._strExcList:
                excStr = excStr + str(item) + '\n<BR>'

        return excStr

class SchainWarning(Exception):
    pass


if __name__ == '__main__':
    
    test = SchainNotify()

    test.sendAlert('This is a message from the python module SchainNotify', 'Test from SchainNotify')

    print('Hopefully message sent - check.')


'''
Created on Aug 1, 2017

@author: Juan C. Espinoza
'''

import os
import sys
import time
import json
import glob
import datetime

import numpy
import h5py

import schainpy.admin
from schainpy.model.io.jroIO_base import LOCALTIME, Reader
from schainpy.model.proc.jroproc_base import ProcessingUnit, Operation, MPDecorator
from schainpy.model.data.jrodata import Parameters
from schainpy.utils import log

try:
    import madrigal.cedar
except:
    pass

try:
    basestring
except:
    basestring = str

DEF_CATALOG = {
    'principleInvestigator': 'Marco Milla',
    'expPurpose': '',
    'cycleTime': '',
    'correlativeExp': '',
    'sciRemarks': '',
    'instRemarks': ''
    }
    
DEF_HEADER = {
    'kindatDesc': '',
    'analyst': 'Jicamarca User',
    'comments': '',
    'history': ''
    }

MNEMONICS = {
    10: 'jro',
    11: 'jbr',
    840: 'jul',
    13: 'jas',
    1000: 'pbr',
    1001: 'hbr',
    1002: 'obr',
    400: 'clr'

}

UT1970 = datetime.datetime(1970, 1, 1) - datetime.timedelta(seconds=time.timezone)

def load_json(obj):
    '''
    Parse json as string instead of unicode
    '''

    if isinstance(obj, str):
        iterable = json.loads(obj)
    else:
        iterable = obj

    if isinstance(iterable, dict):
        return {str(k): load_json(v) if isinstance(v, dict) else str(v) if isinstance(v, basestring) else v
            for k, v in list(iterable.items())}
    elif isinstance(iterable, (list, tuple)):
        return [str(v) if isinstance(v, basestring) else v for v in iterable]
    
    return iterable


class MADReader(Reader, ProcessingUnit):

    def __init__(self):

        ProcessingUnit.__init__(self)

        self.dataOut = Parameters()    
        self.counter_records = 0
        self.nrecords = None
        self.flagNoMoreFiles = 0
        self.filename = None        
        self.intervals = set()
        self.datatime = datetime.datetime(1900,1,1)
        self.format = None
        self.filefmt = "***%Y%m%d*******"
        
    def setup(self, **kwargs):
                
        self.set_kwargs(**kwargs)
        self.oneDDict = load_json(self.oneDDict)
        self.twoDDict = load_json(self.twoDDict)
        self.ind2DList = load_json(self.ind2DList)
        self.independentParam = self.ind2DList[0]

        if self.path is None:
            raise ValueError('The path is not valid')

        self.open_file = open
        self.open_mode = 'rb'

        if self.format is None:
            raise ValueError('The format is not valid choose simple or hdf5')
        elif self.format.lower() in ('simple', 'txt'):
            self.ext = '.txt'
        elif self.format.lower() in ('cedar',):
            self.ext = '.001'
        else:
            self.ext = '.hdf5'
            self.open_file = h5py.File
            self.open_mode = 'r'

        if self.online:
            log.log("Searching files in online mode...", self.name)

            for nTries in range(self.nTries):
                fullpath = self.searchFilesOnLine(self.path, self.startDate,
                    self.endDate, self.expLabel, self.ext, self.walk, 
                    self.filefmt, self.folderfmt)

                try:
                    fullpath = next(fullpath)
                except:
                    fullpath = None
                
                if fullpath:
                    break

                log.warning(
                    'Waiting {} sec for a valid file in {}: try {} ...'.format(
                        self.delay, self.path, nTries + 1), 
                    self.name)
                time.sleep(self.delay)

            if not(fullpath):
                raise schainpy.admin.SchainError(
                    'There isn\'t any valid file in {}'.format(self.path))            
           
        else:
            log.log("Searching files in {}".format(self.path), self.name)
            self.filenameList = self.searchFilesOffLine(self.path, self.startDate, 
                self.endDate, self.expLabel, self.ext, self.walk, self.filefmt, self.folderfmt)
        
        self.setNextFile()

    def readFirstHeader(self):
        '''Read header and data'''

        self.parseHeader()
        self.parseData()
        self.blockIndex = 0
        
        return        

    def parseHeader(self):
        '''
        '''

        self.output = {}
        self.version = '2'
        s_parameters = None
        if self.ext == '.txt':
            self.parameters = [s.strip().lower() for s in self.fp.readline().decode().strip().split(' ') if s]
        elif self.ext == '.hdf5':
            self.metadata = self.fp['Metadata']
            if '_record_layout' in self.metadata:
                s_parameters = [s[0].lower().decode() for s in self.metadata['Independent Spatial Parameters']]
                self.version = '3'
            self.parameters = [s[0].lower().decode() for s in self.metadata['Data Parameters']]

        log.success('Parameters found: {}'.format(self.parameters),
                    'MADReader')
        if s_parameters:
            log.success('Spatial parameters found: {}'.format(s_parameters),
                        'MADReader')
        
        for param in list(self.oneDDict.keys()):
            if param.lower() not in self.parameters:
                log.warning(
                    'Parameter {} not found will be ignored'.format(
                        param),
                    'MADReader')
                self.oneDDict.pop(param, None)
        
        for param, value in list(self.twoDDict.items()):
            if param.lower() not in self.parameters:
                log.warning(
                    'Parameter {} not found, it will be ignored'.format(
                        param),
                    'MADReader')
                self.twoDDict.pop(param, None)
                continue
            if isinstance(value, list):
                if value[0] not in self.output:
                    self.output[value[0]] = []
                self.output[value[0]].append([])

    def parseData(self):
        '''
        '''

        if self.ext == '.txt':
            self.data = numpy.genfromtxt(self.fp, missing_values=('missing'))
            self.nrecords = self.data.shape[0]
            self.ranges = numpy.unique(self.data[:,self.parameters.index(self.independentParam.lower())])
            self.counter_records = 0
        elif self.ext == '.hdf5':
            self.data = self.fp['Data']
            self.ranges = numpy.unique(self.data['Table Layout'][self.independentParam.lower()])
            self.times = numpy.unique(self.data['Table Layout']['ut1_unix'])
            self.counter_records = int(self.data['Table Layout']['recno'][0])
            self.nrecords = int(self.data['Table Layout']['recno'][-1])

    def readNextBlock(self):

        while True:
            self.flagDiscontinuousBlock = 0
            if self.counter_records == self.nrecords:
                self.setNextFile()                

            self.readBlock()
            
            if (self.datatime < datetime.datetime.combine(self.startDate, self.startTime)) or \
               (self.datatime > datetime.datetime.combine(self.endDate, self.endTime)):
                log.warning(
                    'Reading Record No. {}/{} -> {} [Skipping]'.format(
                        self.counter_records,
                        self.nrecords,
                        self.datatime.ctime()),
                    'MADReader')
                continue
            break

        log.log(
            'Reading Record No. {}/{} -> {}'.format(
                self.counter_records,
                self.nrecords,
                self.datatime.ctime()),
            'MADReader')

        return 1

    def readBlock(self):
        '''
        '''
        dum = []
        if self.ext == '.txt':
            dt = self.data[self.counter_records][:6].astype(int)
            if datetime.datetime(dt[0], dt[1], dt[2], dt[3], dt[4], dt[5]).date() > self.datatime.date():
                self.flagDiscontinuousBlock = 1
            self.datatime = datetime.datetime(dt[0], dt[1], dt[2], dt[3], dt[4], dt[5])
            while True:
                dt = self.data[self.counter_records][:6].astype(int)
                datatime = datetime.datetime(dt[0], dt[1], dt[2], dt[3], dt[4], dt[5])
                if datatime == self.datatime:
                    dum.append(self.data[self.counter_records])
                    self.counter_records += 1
                    if self.counter_records == self.nrecords:
                        break
                    continue
                self.intervals.add((datatime-self.datatime).seconds)                
                break
        elif self.ext == '.hdf5':
            datatime = datetime.datetime.utcfromtimestamp(
                self.times[self.counter_records])
            dum = self.data['Table Layout'][self.data['Table Layout']['recno']==self.counter_records]
            self.intervals.add((datatime-self.datatime).seconds)
            if datatime.date()>self.datatime.date():
                self.flagDiscontinuousBlock = 1
            self.datatime = datatime
            self.counter_records += 1            
        
        self.buffer = numpy.array(dum)
        return

    def set_output(self):
        '''
        Storing data from buffer to dataOut object
        '''        

        parameters = [None for __ in self.parameters]

        for param, attr in list(self.oneDDict.items()):            
            x = self.parameters.index(param.lower())
            setattr(self.dataOut, attr, self.buffer[0][x])

        for param, value in list(self.twoDDict.items()):
            dummy = numpy.zeros(self.ranges.shape) + numpy.nan               
            if self.ext == '.txt':
                x = self.parameters.index(param.lower())
                y = self.parameters.index(self.independentParam.lower())            
                ranges = self.buffer[:,y]
                #if self.ranges.size == ranges.size:
                #    continue
                index = numpy.where(numpy.in1d(self.ranges, ranges))[0]
                dummy[index] = self.buffer[:,x]
            else:
                ranges = self.buffer[self.independentParam.lower()]
                index = numpy.where(numpy.in1d(self.ranges, ranges))[0]
                dummy[index] = self.buffer[param.lower()]
            
            if isinstance(value, str):
                if value not in self.independentParam:             
                    setattr(self.dataOut, value, dummy.reshape(1,-1))
            elif isinstance(value, list):                
                self.output[value[0]][value[1]] = dummy
                parameters[value[1]] = param
        for key, value in list(self.output.items()):
            setattr(self.dataOut, key, numpy.array(value))
        
        self.dataOut.parameters = [s for s in parameters if s]
        self.dataOut.heightList = self.ranges
        self.dataOut.utctime = (self.datatime - datetime.datetime(1970, 1, 1)).total_seconds()
        self.dataOut.utctimeInit = self.dataOut.utctime  
        self.dataOut.paramInterval = min(self.intervals)
        self.dataOut.useLocalTime = False        
        self.dataOut.flagNoData = False        
        self.dataOut.nrecords = self.nrecords
        self.dataOut.flagDiscontinuousBlock = self.flagDiscontinuousBlock

    def getData(self):
        '''
        Storing data from databuffer to dataOut object
        '''

        if not  self.readNextBlock():
            self.dataOut.flagNoData = True
            return 0

        self.set_output()

        return 1

    def run(self, **kwargs):

        if not(self.isConfig):
            self.setup(**kwargs)
            self.isConfig = True

        self.getData()

        return

@MPDecorator
class MADWriter(Operation):
    '''Writing module for Madrigal files
    
type: external

Inputs:
            path        path where files will be created
            oneDDict    json of one-dimensional parameters in record where keys
                        are Madrigal codes (integers or mnemonics) and values the corresponding
                        dataOut attribute e.g: {
                            'gdlatr': 'lat',
                            'gdlonr': 'lon',
                            'gdlat2':'lat',
                            'glon2':'lon'}
            ind2DList   list of independent spatial two-dimensional parameters e.g:
                        ['heigthList']
            twoDDict    json of two-dimensional parameters in record where keys
                        are Madrigal codes (integers or mnemonics) and values the corresponding
                        dataOut attribute if multidimensional array specify as tupple
                        ('attr', pos) e.g: {
                            'gdalt': 'heightList',
                            'vn1p2': ('data_output', 0),
                            'vn2p2': ('data_output', 1),
                            'vn3': ('data_output', 2),
                            'snl': ('data_SNR', 'db')
                            }
            metadata    json of madrigal metadata (kinst, kindat, catalog and header)
            format      hdf5, cedar
            blocks      number of blocks per file'''

    __attrs__ = ['path', 'oneDDict', 'ind2DList', 'twoDDict','metadata', 'format', 'blocks']
    missing = -32767
    
    def __init__(self):

        Operation.__init__(self)
        self.dataOut = Parameters()
        self.counter = 0
        self.path = None
        self.fp = None

    def run(self, dataOut, path, oneDDict, ind2DList='[]', twoDDict='{}',
            metadata='{}', format='cedar', **kwargs):
        
        if not self.isConfig:
            self.setup(path, oneDDict, ind2DList, twoDDict, metadata, format, **kwargs)
            self.isConfig = True
        
        self.dataOut = dataOut        
        self.putData() 
        return 1
    
    def setup(self, path, oneDDict, ind2DList, twoDDict, metadata, format, **kwargs):
        '''
        Configure Operation        
        '''
                
        self.path = path
        self.blocks = kwargs.get('blocks', None)
        self.counter = 0
        self.oneDDict = load_json(oneDDict)
        self.twoDDict = load_json(twoDDict)
        self.ind2DList = load_json(ind2DList)
        meta = load_json(metadata)        
        self.kinst = meta.get('kinst')
        self.kindat = meta.get('kindat')
        self.catalog = meta.get('catalog', DEF_CATALOG)
        self.header = meta.get('header', DEF_HEADER)
        if format == 'cedar':
            self.ext = '.dat'
            self.extra_args = {}
        elif format == 'hdf5':
            self.ext = '.hdf5'
            self.extra_args = {'ind2DList': self.ind2DList}
        
        self.keys = [k.lower() for k in self.twoDDict]        
        if 'range' in self.keys:
            self.keys.remove('range')
        if 'gdalt' in self.keys:
            self.keys.remove('gdalt')

    def setFile(self):
        '''
        Create new cedar file object
        '''

        self.mnemonic = MNEMONICS[self.kinst]   #TODO get mnemonic from madrigal
        date = datetime.datetime.utcfromtimestamp(self.dataOut.utctime)

        filename = '{}{}{}'.format(self.mnemonic,
                                   date.strftime('%Y%m%d_%H%M%S'),
                                   self.ext)
       
        self.fullname = os.path.join(self.path, filename)
    
        if os.path.isfile(self.fullname) : 
            log.warning(
                'Destination file {} already exists, previous file deleted.'.format(
                    self.fullname),
                'MADWriter')
            os.remove(self.fullname)
        
        try:
            log.success(
                'Creating file: {}'.format(self.fullname),
                'MADWriter')
            if not os.path.exists(self.path):
                os.makedirs(self.path)
            self.fp = madrigal.cedar.MadrigalCedarFile(self.fullname, True)
        except ValueError as e:
            log.error(
                'Impossible to create a cedar object with "madrigal.cedar.MadrigalCedarFile"',
                'MADWriter')
            return

        return 1

    def writeBlock(self):
        '''
        Add data records to cedar file taking data from oneDDict and twoDDict
        attributes.
        Allowed parameters in: parcodes.tab
        '''

        startTime = datetime.datetime.utcfromtimestamp(self.dataOut.utctime)
        endTime = startTime + datetime.timedelta(seconds=self.dataOut.paramInterval)
        heights = self.dataOut.heightList

        if self.ext == '.dat':
            for key, value in list(self.twoDDict.items()):
                if isinstance(value, str):
                    data = getattr(self.dataOut, value)
                    invalid = numpy.isnan(data)
                    data[invalid] = self.missing
                elif isinstance(value, (tuple, list)):
                    attr, key = value
                    data = getattr(self.dataOut, attr)
                    invalid = numpy.isnan(data)
                    data[invalid] = self.missing

        out = {}
        for key, value in list(self.twoDDict.items()):
            key = key.lower()
            if isinstance(value, str):
                if 'db' in value.lower():
                    tmp = getattr(self.dataOut, value.replace('_db', ''))
                    SNRavg = numpy.average(tmp, axis=0)
                    tmp = 10*numpy.log10(SNRavg)
                else:
                    tmp = getattr(self.dataOut, value)
                out[key] = tmp.flatten()[:len(heights)]
            elif isinstance(value, (tuple, list)):
                attr, x = value
                data = getattr(self.dataOut, attr)                
                out[key] = data[int(x)][:len(heights)]
        
        a = numpy.array([out[k] for k in self.keys])
        nrows = numpy.array([numpy.isnan(a[:, x]).all() for x in range(len(heights))])
        index = numpy.where(nrows == False)[0]

        rec = madrigal.cedar.MadrigalDataRecord(
            self.kinst,
            self.kindat,
            startTime.year,
            startTime.month,
            startTime.day,
            startTime.hour,
            startTime.minute,
            startTime.second,
            startTime.microsecond/10000,
            endTime.year,
            endTime.month,
            endTime.day,
            endTime.hour,
            endTime.minute,
            endTime.second,
            endTime.microsecond/10000,
            list(self.oneDDict.keys()),
            list(self.twoDDict.keys()),
            len(index),
            **self.extra_args
        )

        # Setting 1d values        
        for key in self.oneDDict:
            rec.set1D(key, getattr(self.dataOut, self.oneDDict[key]))

        # Setting 2d values
        nrec = 0
        for n in index:            
            for key in out:
                rec.set2D(key, nrec, out[key][n])
            nrec += 1 

        self.fp.append(rec)
        if self.ext == '.hdf5' and self.counter % 500 == 0 and self.counter > 0:
            self.fp.dump()
        if self.counter % 20 == 0 and self.counter > 0:
            log.log(
                'Writing {} records'.format(
                    self.counter),
                'MADWriter')

    def setHeader(self):
        '''
        Create an add catalog and header to cedar file
        ''' 
        
        log.success('Closing file {}'.format(self.fullname), 'MADWriter')

        if self.ext == '.dat':
            self.fp.write()
        else:
            self.fp.dump()
            self.fp.close()
        
        header = madrigal.cedar.CatalogHeaderCreator(self.fullname)        
        header.createCatalog(**self.catalog)
        header.createHeader(**self.header)
        header.write()
              
    def putData(self):

        if self.dataOut.flagNoData:
            return 0        
        
        if self.dataOut.flagDiscontinuousBlock or self.counter == self.blocks:
            if self.counter > 0:
                self.setHeader()
            self.counter = 0

        if self.counter == 0:
            self.setFile()
        
        self.writeBlock()
        self.counter += 1        
        
    def close(self):
        
        if self.counter > 0:                
            self.setHeader()